import os
import sys

from argparse import ArgumentParser
import torch
import wandb

from signaturesnet.trainers import train_classifier
from signaturesnet.utilities.io import update_dict, read_config, write_result

DEFAULT_CONFIG_FILE = ["configs/classifier/classifier.yaml"]

if __name__ == "__main__":
    # Parse command-line arguments
    parser = ArgumentParser()
    parser.add_argument(
        '--config_file', action='store', nargs=1, type=str, required=False, default=DEFAULT_CONFIG_FILE,
        help=f'Path to yaml file containing all needed parameters.\nThey can be overwritten by command-line arguments'
    )

    parser.add_argument(
        '--model_id', action='store', nargs=1, type=str, required=False,
        help=f'Unique id given to the trained model.'
    )

    # Data args
    parser.add_argument(
        '--sigmoid_params', action='store', nargs=1, type=float, required=False,
        help='Sigmoid parameters for normalization.'
    )

    # Train args
    parser.add_argument(
        '--batch_size', action='store', nargs=1, type=int, required=False,
        help='Training batch size.'
    )
    parser.add_argument(
        '--lr', action='store', nargs=1, type=float, required=False,
        help='Learning rate.'
    )
    
    # Model args
    parser.add_argument(
        '--num_neurons', action='store', nargs=1, type=int, required=False,
        help='Neurons of the positive error network.'
    )
    parser.add_argument(
        '--num_hidden_layers', action='store', nargs=1, type=int, required=False,
        help='Layers of the positive error network.'
    )

    _args = parser.parse_args()

    # Read & config
    config = read_config(path=getattr(_args, "config_file")[0])
    config = update_dict(config=config, args=_args)
    
    print("Using config:", config)
    score = train_classifier(config=config)
    write_result(score, "../tmp/classifier_score_%s.txt"%config["model_id"])



import os
import sys
import logging

import numpy as np
import random
import torch
from tqdm import tqdm

from signaturesnet.utilities.io import csv_to_tensor, read_model

class DataGenerator:

    def __init__(self,
                 signatures,
                 seed=None,
                 normalize=True,
                 shuffle=True):
        self.signatures = signatures
        self.total_signatures = signatures.shape[1]
        self.normalize = normalize
        self.shuffle = shuffle
        self._set_all_seeds(seed)

        logging.info("Total signatures: %s"%self.total_signatures)


    def _set_all_seeds(self, seed):
        if seed is None:
            return
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = True


    def _get_nummuts(self, split, large_or_low, size):
        assert split in ["train", "valid", "test"]
        assert large_or_low in ["large", "low", "superlow"]

        # For train and validation we sample from ranges
        if split == "train" or split == "valid":
            if large_or_low == "superlow": 
                range_muts = [1, 5, 10, 15, 20, 25]
                n_ranges = len(range_muts) - 1
                num_samples = int(size/n_ranges)
                ind_range_muts = [0]*num_samples + [1]*num_samples + [2]*num_samples + \
                    [3]*num_samples + [4]*(num_samples + size%n_ranges)

            elif large_or_low == 'low':
                range_muts = [15, 50, 100, 250, 500, 1000, 5000, 1e4, 1e5]
                n_ranges = len(range_muts) - 1
                num_samples = int(size/n_ranges)
                ind_range_muts = [0]*num_samples + [1]*num_samples + [2]*num_samples + \
                    [3]*num_samples + [4]*num_samples + [5]*num_samples + [6]*num_samples +\
                    [7]*(num_samples + size%n_ranges)
            
            elif large_or_low == 'large':
                range_muts = [1e3, 5e3, 1e4, 5e4, 1e5, 5e5]
                n_ranges = len(range_muts) - 1
                num_samples = int(size/n_ranges)
                ind_range_muts = [0]*num_samples + [1]*num_samples + \
                    [2]*num_samples + [3]*num_samples + [-1]*(num_samples + size%8)
                # -1 means real distribution

            nummuts = [np.random.randint(range_muts[ind_range_muts[i]], range_muts[ind_range_muts[i] + 1])\
                        if ind_range_muts[i] != -1 else -1 for i in range(len(ind_range_muts))]
            return nummuts

        # For testing we use exact number of mutations
        num_samples = int(size/10.0)
        num_muts = [25]*num_samples + [50]*num_samples + [100]*num_samples +\
                   [250]*num_samples + [500]*num_samples + [1e3]*num_samples +\
                   [5e3]*num_samples + [1e4]*num_samples + [5e4]*num_samples +\
                   [1e5]*(num_samples + size%10)
        return num_muts


    def _sample_from_sig(self, signature, num_mut):
        sample = signature
        if num_mut > 0 and num_mut < 1e5:  # only do sampling for num_mut < 1e5
            c = torch.distributions.categorical.Categorical(probs=signature)
            sample_shape = torch.Size([num_mut, ])
            samples = c.sample(sample_shape=sample_shape).type(torch.float32)
            n_bins = signature.shape[0]
            sample = torch.histc(samples, bins=n_bins, min=0, max=n_bins - 1)
            if self.normalize:
                sample = sample/float(num_mut)
        else:
            sample = sample/torch.sum(sample)
        return sample


    def make_input(self, labels, split, large_low, augmentation=1, nummuts=None):
        """Create a labelled dataset of mutation vectors
           from a tensor of labels.
           Returns:
            inputs (mutational vector)
            labels (including an appended column with the number of mutations used)
        """
        labels = torch.cat([labels]*augmentation, dim = 0).to(torch.float32)
        if nummuts is None:
            nummuts = self._get_nummuts(split, large_low, size=labels.shape[0])
        else:
            nummut_means = torch.cat([nummuts]*augmentation, dim=0)
            nummut_stds = nummut_means/10.0
            nummuts = torch.clip(torch.distributions.normal.Normal(nummut_means, nummut_stds).sample().type(torch.int64), 10, 1e5)

        input_batch = torch.empty((labels.shape[0], 96))
        labels_batch = torch.empty((labels.shape[0], self.total_signatures + 1))

        logging.info("Sampling from labels...")
        for i, num_mut in tqdm(enumerate(nummuts)):
            # Compute resulting signature
            signature = torch.einsum("ij,j->i", (self.signatures, labels[i]))

            # Sample the mutational vector from the given distribution
            sample = self._sample_from_sig(signature=signature,
                                           num_mut=int(num_mut))
            
            # Store Values
            # For the real distribution we say we have more than 1e5 mutations
            num_mut = float(num_mut) if num_mut != -1 else float(np.random.randint(1e5, 1e6))
            input_batch[i, :] = sample
            labels_batch[i, :] = torch.cat([labels[i, :], torch.tensor([float(num_mut)])])
        return input_batch, labels_batch


    def make_random_set(self,
                        split,
                        large_low,
                        num_samples,
                        min_n_signatures=1,
                        max_n_signatures=10,
                        min_weight=0.1):
        """Create a labelled dataset of mutation vectors
           randomly combining the signatures.
        """
        assert(min_n_signatures > 0 and max_n_signatures <= self.total_signatures)

        labels = torch.empty((num_samples, self.total_signatures))

        logging.info("Generating random labels...")
        for i in tqdm(range(num_samples)):
            if i == 1:
                print("here")
            # Pick the number of signatures involved
            n_signatures = np.random.randint(min_n_signatures, max_n_signatures + 1)

            # Select n_signatures
            signature_ids = torch.randperm(self.total_signatures)[:n_signatures]

            # Assign weights randomly
            weights = torch.rand(size=(n_signatures,))
            weights = weights/torch.sum(weights)
            weights[weights < min_weight] = 0
            weights = weights/torch.sum(weights)
            label = torch.zeros(self.total_signatures).scatter_(
                dim=0, index=signature_ids, src=weights)

            # Store
            labels[i, :] = label
        return self.make_input(labels, split, large_low)


    def make_realistic_set(self,
                           generator_model_path,
                           split,
                           large_low,
                           num_samples,
                           std = 1.0):
        """Create a labelled dataset of mutation vectors
           from the generator output weights.
        """
        generator = read_model(generator_model_path, device="cpu")
        logging.info("Generating realistic labels...")
        labels = generator.generate(num_samples, std = std)
        return self.make_input(labels, split, large_low)


import collections
import os
import sys

import numpy as np
import torch
import torch.optim as optim
from torch.utils.data import DataLoader
from tqdm import tqdm
import wandb

from signaturesnet.loggers.finetuner_logger import FinetunerLogger
from signaturesnet.models import FineTunerLowNumMut, FineTunerLargeNumMut
from signaturesnet.utilities.metrics import get_fp_fn_soft, get_classification_metrics, get_kl_divergence
from signaturesnet.utilities.io import save_model
from signaturesnet.utilities.temporal_io import read_data_final_finetuner

class FinetunerTrainer:
    def __init__(self,
                 iterations,
                 train_data,
                 val_data,
                 network_type,
                 sigmoid_params=[5000, 2000],
                 loging_path="../runs",
                 num_classes=72,
                 log_freq=100,
                 model_path=None,  # File where to save model learned weights None to not save
                 device=torch.device("cuda:0")):
        self.iterations = iterations  # Now iteration refers to passes through all dataset
        self.num_classes = num_classes
        self.sigmoid_params = sigmoid_params
        self.device = device
        self.log_freq = log_freq
        self.model_path = model_path
        self.train_dataset = train_data
        self.val_dataset = val_data
        self.network_type = network_type
        assert (self.network_type in ['low', 'large'])
        self.logger = FinetunerLogger()

    def __loss(self, prediction, label, FP, FN):
        # if self.network_type == 'low':
        #     l = get_kl_divergence(predicted_label=prediction, true_label=label)
        # if self.network_type == 'large':
        # l = get_jensen_shannon(predicted_label=prediction, true_label=label)
        l = get_kl_divergence(predicted_label=prediction, true_label=label)
        l += FP / prediction.shape[0]
        return l

    def objective(self,
                  batch_size,
                  lr,
                  num_hidden_layers,
                  num_units,
                  plot=False):

        print(batch_size, lr, num_hidden_layers, num_units)

        dataloader = DataLoader(dataset=self.train_dataset,
                                batch_size=int(batch_size),
                                shuffle=True)
        model = None
        if self.network_type == "low":
            model = FineTunerLowNumMut(num_classes=self.num_classes,
                                       num_hidden_layers=int(num_hidden_layers),
                                       num_units=int(num_units),
                                       sigmoid_params=self.sigmoid_params)
        elif self.network_type == "large":
            model = FineTunerLargeNumMut(num_classes=self.num_classes,
                                       num_hidden_layers=int(num_hidden_layers),
                                       num_units=int(num_units),
                                       sigmoid_params=self.sigmoid_params)
        # model = read_model(self.model_path)
        model.to(self.device)

        # if plot:
        #     wandb.watch(model, log_freq=self.log_freq, log_graph=True)

        optimizer = optim.Adam(model.parameters(), lr=lr, weight_decay=1e-4)

        l_vals = collections.deque(maxlen=50)
        max_found = -np.inf
        step = 0
        for _ in range(self.iterations):
            for train_input, train_label, train_baseline, num_mut, _ in tqdm(dataloader):
                model.train()  # NOTE: Very important! Otherwise we zero the gradient
                optimizer.zero_grad()   
                if self.network_type == "large":             
                    train_prediction = model(train_input, train_baseline, num_mut)
                else:
                    train_prediction = model(train_input, num_mut)

                train_FP, train_FN = get_fp_fn_soft(label_batch=train_label,
                                                    prediction_batch=train_prediction)
                # train_FP, train_FN = None, None
                train_loss = self.__loss(prediction=train_prediction,
                                         label=train_label,
                                         FP=train_FP,
                                         FN=train_FN)

                train_loss.backward()
                optimizer.step()

                model.eval()
                with torch.no_grad():
                    train_classification_metrics = get_classification_metrics(label_batch=train_label,
                                                                              prediction_batch=train_prediction)
                    if self.network_type == "large":     
                        val_prediction = model(self.val_dataset.inputs, self.val_dataset.prev_guess, self.val_dataset.num_mut)
                    else:
                        val_prediction = model(self.val_dataset.inputs, self.val_dataset.num_mut)
                    # val_FP, val_FN = None, None
                    val_FP, val_FN = get_fp_fn_soft(label_batch=self.val_dataset.labels,
                                                    prediction_batch=val_prediction)
                    val_loss = self.__loss(prediction=val_prediction,
                                           label=self.val_dataset.labels,
                                           FP=val_FP,
                                           FN=val_FN)
                    l_vals.append(val_loss.item())
                    max_found = max(max_found, -np.nanmean(l_vals))

                if plot and step % self.log_freq == 0:
                    val_classification_metrics = get_classification_metrics(label_batch=self.val_dataset.labels,
                                                                            prediction_batch=val_prediction)
                    self.logger.log(train_loss=train_loss,
                                    train_prediction=train_prediction,
                                    train_label=train_label,
                                    train_classification_metrics=train_classification_metrics,
                                    val_loss=val_loss,
                                    val_prediction=val_prediction,
                                    val_label=self.val_dataset.labels,
                                    val_classification_metrics=val_classification_metrics,
                                    step=step)

                # if self.model_path is not None and step % 1000 == 0:
                #     save_model(model=model, directory=self.model_path + '_it' + str(step))
                step += 1
        save_model(model=model, directory=self.model_path)
        return max_found


def train_finetuner(config, data_folder="../data", name=None, train_data=None, val_data=None) -> float:
    import os
    # Select training device
    dev = "cuda" if config["device"] == "cuda" and torch.cuda.is_available(
    ) else "cpu"
    device = torch.device(dev)
    print("Using device:", device)

    # Set paths
    finetuner_path = os.path.join(config["models_dir"], config["model_id"])

    if config["enable_logging"]:
        run = wandb.init(project=config["wandb_project_id"],
                         entity='sig-net',
                         config=config,
                         name=config["model_id"] if name is None else name)

    load_data = config["load_data"]
    assert (train_data is None and val_data is None) or (train_data is not None and val_data is not None)
    if train_data is None or val_data is None:
        # train_data, val_data = read_data(experiment_id=config["data_id"],
        #                                 source=config["source"],
        #                                 data_folder=data_folder,
        #                                 device=dev)

        train_data, val_data = read_data_final_finetuner(device = dev,
                                                         data_id = config["data_id"],
                                                         data_folder = "../data/", 
                                                         network_type = config["network_type"])
    
    trainer = FinetunerTrainer(iterations=config["iterations"],  # Passes through all dataset
                               train_data=train_data,
                               val_data=val_data,
                               network_type=config["network_type"],
                               num_classes=config["num_classes"],
                               sigmoid_params=config["sigmoid_params"],
                               device=device,
                               model_path=finetuner_path)

    min_val = trainer.objective(batch_size=config["batch_size"],
                                lr=config["lr"],
                                num_hidden_layers=config["num_hidden_layers"],
                                num_units=config["num_neurons"],
                                plot=True)
    if config["enable_logging"]:
        wandb.log({"validation_score": min_val})
        run.finish()
    return min_val



import copy

import numpy as np
from scipy.stats import wasserstein_distance
import torch
import torch.nn as nn


# USED IN CLASSIFIER
def accuracy(prediction, label):
    assert(prediction.shape == label.shape)
    assert(prediction.dtype == torch.int64)
    assert(label.dtype == torch.int64)
    return (torch.sum(prediction == label).float()/torch.numel(prediction))*100.

def false_realistic(prediction, label):
    assert(prediction.shape == label.shape)
    assert(prediction.dtype == torch.int64)
    # assert(label.dtype == torch.int64)
    # print(label[prediction == 1])
    # print(torch.sum(label[prediction == 1] == 0))
    return torch.true_divide(torch.sum(label[prediction == 1] == 0), (torch.numel(label)))*100 

def false_random(prediction, label):
    assert(prediction.shape == label.shape)
    assert(prediction.dtype == torch.int64)
    # assert(label.dtype == torch.int64)
    return torch.true_divide(torch.sum(label[prediction == 0] == 1), (torch.numel(label)))*100 

# USED IN FINE TUNER
def get_MSE(predicted_label, true_label):
    return torch.nn.MSELoss()(predicted_label, true_label)

def get_cosine_similarity(predicted_label, true_label, dim=None):
    predicted_label = torch.nn.functional.normalize(
        predicted_label, dim=1, p=2)
    true_label = torch.nn.functional.normalize(true_label, dim=1, p=2)
    if dim is not None:
        return torch.abs(torch.einsum("ij,ij->i", (predicted_label, true_label)))
    return torch.mean(torch.abs(torch.einsum("ij,ij->i", (predicted_label, true_label))))

def get_negative_cosine_similarity(predicted_label, true_label):
    return -get_cosine_similarity(predicted_label, true_label)

def get_entropy(predicted_label):
    entropy = torch.mean(torch.distributions.categorical.Categorical(
        probs=predicted_label.cpu().detach()).entropy())
    return entropy

def get_cross_entropy2(predicted_label, true_label):
    #predicted_label_local = copy.deepcopy(predicted_label)
    #predicted_label_local += 1e-6
    return torch.mean(-torch.einsum("ij,ij->i", (true_label, torch.log(predicted_label))))

def get_kl_divergence(predicted_label, true_label):
    _EPS = 1e-9
    predicted_label = predicted_label + _EPS
    return get_cross_entropy2(predicted_label, true_label) - get_entropy(true_label)

def get_jensen_shannon(predicted_label, true_label):
    _EPS = 1e-6
    r = (predicted_label + true_label)/2 + _EPS
    term1 = torch.mean(torch.einsum("ij,ij->i",(predicted_label, torch.log(torch.div(predicted_label + _EPS, r)))))
    term2 = torch.mean(torch.einsum("ij,ij->i",(true_label, torch.log(torch.div(true_label + _EPS, r)))))
    return 0.5 * (term1 + term2) 

def get_wasserstein_distance(predicted_label, true_label):
    dist = 0
    for i in range(predicted_label.shape[0]):
        dist += wasserstein_distance(
            predicted_label[i, :].cpu().detach().numpy(), true_label[i, :].cpu().detach().numpy())
    return torch.from_numpy(np.array(dist/predicted_label.shape[0]))

def get_classification_metrics(label_batch, prediction_batch, cutoff=0.01):
    batch_size = float(label_batch.shape[0])
    kl = get_kl_divergence(prediction_batch, label_batch)
    label_mask = (label_batch > cutoff).type(torch.int).float()
    prediction_mask = (prediction_batch > cutoff).type(torch.int).float()
    fp = torch.sum(label_mask - prediction_mask < -0.1)
    fn = torch.sum(label_mask - prediction_mask > 0.1)
    tp = torch.sum(torch.einsum("bi,bi->b", prediction_mask, label_mask))
    tn = torch.sum(torch.einsum("bi,bi->b", 1 - prediction_mask, 1 - label_mask))
    fpr = fp/(tn + fp)
    fnr = fn/(tp + fn)
    sensitivity = tp/torch.sum(label_mask)
    specificity = tn/torch.sum(1 - label_mask)
    accuracy = (tp + tn)/(batch_size*label_batch.shape[1])
    precision = tp / (tp + fp)
    mae = torch.abs(label_batch - prediction_batch)
    mae_by_sign = torch.mean(torch.abs(label_batch - prediction_batch), dim=0)
    MAE_p = torch.sum(torch.einsum("bi,bi->b", mae, label_mask))/(tp + fn)
    MAE_n = torch.sum(torch.einsum("bi,bi->b", mae, 1 - label_mask))/(tn + fp)
    # Q95_p = torch.quantile(mae[label])
    return {"fpr": fpr*100, "fnr": fnr*100,
            "sens: tp/p %": sensitivity*100., "spec: tn/n %": specificity*100., "accuracy %": accuracy*100., "precision %": precision*100.,
            "MAE_p": MAE_p, "MAE_n": MAE_n, "MAE":mae.mean(), "KL":kl, "MAE_sign":mae_by_sign.detach()}

def get_fp_fn_soft(label_batch, prediction_batch, cutoff=0.05, softness=100):
    label_mask = nn.Sigmoid()((label_batch - cutoff)*softness)  # ~0 if under cutoff, ~1 if over cutoff (element-wise)
    prediction_mask = nn.Sigmoid()((prediction_batch - cutoff)*softness)  # ~0 if under cutoff, ~1 if over cutoff (element-wise)
    fp = torch.sum(nn.ReLU()(prediction_mask - label_mask))  # only count when pred ~= 1 and label ~= 0
    fn = torch.sum(nn.ReLU()(label_mask - prediction_mask))  # only count when label ~= 1 and pred ~= 0
    return fp, fn

def probs_batch_to_sigs(label_batch, prediction_batch, cutoff=0.05, num_classes=72):
    label_sigs_list = torch.zeros(0, dtype=torch.long)
    predicted_sigs_list = torch.zeros(0, dtype=torch.long)
    for i in range(len(label_batch)):
        for j in range(len(label_batch[i])):
            if label_batch[i][j] > cutoff and prediction_batch[i][j] > cutoff:
                label_sigs_list = torch.cat(
                    [label_sigs_list, torch.from_numpy(np.array([j]))])
                predicted_sigs_list = torch.cat(
                    [predicted_sigs_list, torch.from_numpy(np.array([j]))])
            if label_batch[i][j] > cutoff and prediction_batch[i][j] < cutoff:
                label_sigs_list = torch.cat(
                    [label_sigs_list, torch.from_numpy(np.array([j]))])
                predicted_sigs_list = torch.cat(
                    [predicted_sigs_list, torch.from_numpy(np.array([num_classes]))])
            if label_batch[i][j] < cutoff and prediction_batch[i][j] > cutoff:
                label_sigs_list = torch.cat(
                    [label_sigs_list, torch.from_numpy(np.array([num_classes]))])
                predicted_sigs_list = torch.cat(
                    [predicted_sigs_list, torch.from_numpy(np.array([j]))])
    return label_sigs_list, predicted_sigs_list

# USED IN ERROR FINDER 
def distance_to_interval(label, pred_lower, pred_upper):
    batch_size = float(pred_lower.shape[0])
    lower = label - pred_lower
    upper = pred_upper - label
    lower = nn.ReLU()(-lower)
    upper = nn.ReLU()(-upper)
    dist = lower + upper
    return dist

def interval_width(pred_lower, pred_upper):
    return abs(pred_upper-pred_lower)

def get_pi_metrics_by_sig(label, pred_lower, pred_upper, dim=0):
    """Used to compare prediction interval guesses.
    
    Returns:
        in_prop [float]: Proportion of labels in (pred_lower, pred_upper)
        mean_interval_width [float]: Mean width of the intervals
    """
    EPS = 1e-6
    result = {}
    k_hu = (label <= pred_upper).type(torch.float)  # 1 if label < upper; else 0
    k_hl = (pred_lower <= label).type(torch.float)  # 1 if lower < label; else 0
    k_h = torch.einsum("be,be->be", k_hl, k_hu)  # 1 if label in (lower, upper) else 0
    result["in_prop"] = torch.mean(k_h, dim)  # Hard Prediction Interval Coverage Probability
    result["in_prop_present"] = torch.masked_select(k_h, label > 0.5).mean()  # This flattens all values into a 1-d tensor
    result["in_prop_absent"] = torch.masked_select(k_h, label <= 0.5).mean()

    pred_upper[pred_upper!=pred_upper] = 0
    pred_lower[pred_lower!=pred_lower] = 0
    interval_width = torch.max(torch.zeros_like(label), pred_upper - pred_lower)
    result["mean_interval_width"] = torch.mean(interval_width, dim)
    interval_width_present = torch.masked_select(interval_width, label > EPS)
    interval_width_absent = torch.masked_select(interval_width, label <= EPS)
    result["mean_interval_width_present"] = interval_width_present.mean()
    result["mean_interval_width_absent"] = interval_width_absent.mean()
    return result

def get_pi_metrics(label, pred_lower, pred_upper, collapse=True, dim=0):
    metrics_by_signature = get_pi_metrics_by_sig(label, pred_lower, pred_upper, dim=dim)
    if collapse:
        for key in metrics_by_signature:
            metrics_by_signature[key] = metrics_by_signature[key].mean()
    return metrics_by_signature

def get_soft_qd_loss(label, pred_lower, pred_upper, conf=0.01, lagrange_mult=1e-4, softening_factor=500.0):
    """Used to optimize:
    
    min pred_upper - pred_lower
    s.t. p(label in (pred_lower, pred_upper)) >= 1 - conf

    Algorithm 1 of https://arxiv.org/pdf/1802.07167.pdf

    conf [float]: Interval confidence, proportion of missplaced points which is "ok" usually ~0.05
    lagrange_mult [float]: "How much do we care about missclassifications"
                            Larger lagrange_mult means more correctly guessed points at the expense of larger intervals
    softening_factor [float]: The bigger the closer it is to the real function
                              This means better guesses but harder optimization (less differentiable)
    """
    EPS_ = 1e-9
    EPS2_ = 1e-3

    # Hard in-between constrain
    with torch.no_grad():
        k_hu = (label <= pred_upper).type(torch.float)  # 1 if label <= upper; else 0
        k_hl = (pred_lower <= label).type(torch.float)  # 1 if lower <= label; else 0
        k_h = torch.einsum("be,be->be", k_hl, k_hu)  # 1 if label in (lower, upper) else 0
        PICP_h = torch.mean(k_h)  # Prediction Interval Coverage Probability

    # Softened in-between constrain (same as before but differentiable)
    k_su = nn.Sigmoid()((pred_upper - label + EPS2_)*softening_factor)
    k_sl = nn.Sigmoid()((label - pred_lower + EPS2_)*softening_factor)
    k_s = torch.einsum("be,be->be", k_sl, k_su)
    PICP_s = torch.mean(k_s)  # Soft Prediction Interval Coverage Probability

    # Compute Mean Prediction Interval Width (MPIW)
    MPIW = torch.sum(torch.einsum("be,be->be", (pred_upper - pred_lower), k_h))/(torch.sum(k_h) + EPS_)

    # Compute constrain
    n = float(torch.numel(label))  # number elements in the input
    constrain = (n/(conf*(1.0 - conf)))*torch.max(torch.tensor(0).to(PICP_s), (1.0 - conf) - PICP_s)**2
    
    # Compute and return loss
    loss = MPIW + lagrange_mult*constrain
    return loss, PICP_h, MPIW


def get_entropy(data):
    """Return average entropy across 
    """
    _EPS = 1e-9
    return -torch.sum(data * torch.log(data + _EPS))/data.shape[0]

def get_std(data):
    """Return average std across 
    """
    return torch.mean(torch.std(data, axis=1))

def get_present_sigs(data):
    """Return average std across 
    """
    return torch.mean(torch.count(data, axis=1))

def get_reconstruction_error(mutation_dist, guess, signatures):
    reconstruction = torch.einsum("ij,bj->bi", (signatures, guess))
    errors = torch.sum(torch.nn.MSELoss(reduction='none')(reconstruction, mutation_dist), dim=1)
    return errors

def sets_distances(real, fake):
    """Distances between points of two sets
    """
    def min_dist(point_set, point):
        """Minimum distance between a set and a point
        """
        return torch.sqrt(((point_set - point).pow(2)).mean(dim=1).min())
    with torch.no_grad():
        real_distances = torch.tensor([min_dist(fake, p) for p in real])
        fake_distances = torch.tensor([min_dist(real, p) for p in fake])
    return real_distances, fake_distances

def prop_distances(real, fake):
    """Distances between the proportion of signatures in the real and fake samples
    """
    real_changed = real.detach().clone()
    fake_changed = fake.detach().clone()

    real_changed[real_changed<0.01] = 0
    real_changed[real_changed>=0.01] = 1
    prop_tumors_real = torch.sum(real_changed, dim=0)/real_changed.shape[0]

    fake_changed[fake_changed<0.01] = 0
    fake_changed[fake_changed>=0.01] = 1
    prop_tumors_fake = torch.sum(fake_changed, dim=0)/fake_changed.shape[0]
        
    se_prop = ((prop_tumors_fake-prop_tumors_real)**2).detach().numpy()
    mse_prop = np.mean(se_prop)
    return se_prop, mse_prop

def get_distances_metrics(distances, quantiles=[0.99]):
    metrics = {}
    metrics["mean"] = np.format_float_scientific(torch.mean(distances).item(), precision = 2, exp_digits=1)
    metrics["median"] = np.format_float_scientific(torch.median(distances).item(), precision = 2, exp_digits=1)
    metrics["max"] = np.format_float_scientific(torch.max(distances).item(), precision = 2, exp_digits=1)
    quantiles = torch.quantile(distances, torch.tensor(quantiles), keepdim=True)
    metrics["quantiles"] = np.round(quantiles.detach().cpu().numpy(), 3)
    return metrics

METRICS_DICT = {
    "mse" : get_MSE,
    "cos" : get_negative_cosine_similarity,
    "cross_ent" : get_cross_entropy2,
    "KL" : get_kl_divergence,
    "JS" : get_jensen_shannon,
    "W" : get_wasserstein_distance,
}

if __name__ == "__main__":
    torch.seed = 0

    losses = []
    mpiws = []
    picps = []
    lower = torch.tensor([[0.1]])
    label = torch.tensor([[0.25]])
    u_s = [u/100. for u in range(100)]
    for u in u_s:
        upper = torch.tensor([[u]])

        # in_prop, mean_interval_width = get_pi_metrics(label, lower, upper)
        # print("in_prop", in_prop)
        # print("mean_interval_width", mean_interval_width)

        # loss, picp, mpiw = get_soft_qd_loss(label, lower, upper, conf=0.05, lagrange_mult=0.5, softening_factor=10)
        loss, picp, mpiw = distance_to_interval(label, lower, upper, lagrange_mult=1.0)
        # print("loss", loss)
        losses.append(loss)
        picps.append(picp)
        mpiws.append(mpiw)

    import matplotlib.pyplot as plt
    plt.plot(u_s, losses, label="loss")
    plt.plot(u_s, picps, label="in_prop")
    plt.plot(u_s, mpiws, label="width")
    plt.legend()
    plt.show()


import os
import sys

import torch
import wandb

from signaturesnet.utilities.metrics import accuracy, false_random, false_realistic


class ClassifierLogger:
    def __init__(self):
        self.threshold = 0.5

    def log(self,
            train_loss,
            train_prediction,
            train_label,
            val_loss,
            val_prediction,
            val_label,
            step):

        wandb.log({"train_loss": train_loss})
        wandb.log({"val_loss": val_loss})

        wandb.log({"train_accuracy": accuracy(train_prediction, train_label).item()})
        wandb.log({"val_accuracy": accuracy(val_prediction, val_label).item()})

        wandb.log({"train_false_realistic": false_realistic(train_prediction, train_label).item()})
        wandb.log({"val_false_realistic": false_realistic(val_prediction, val_label).item()})

        wandb.log({"train_false_random": false_random(train_prediction, train_label).item()})
        wandb.log({"val_false_random": false_random(val_prediction, val_label).item()})




import collections
import copy
import os
import pathlib
import sys

import numpy as np
import pandas as pd

import torch
from torch import linalg
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
import wandb

from signaturesnet.utilities.data_partitions import DataPartitions
from signaturesnet.utilities.io import save_model, read_model
from signaturesnet.utilities.metrics import distance_to_interval, get_pi_metrics
from signaturesnet.models import ErrorFinder
from signaturesnet.loggers import ErrorFinderLogger
from signaturesnet.modules import CombinedFinetuner
from signaturesnet.utilities.plotting import plot_error_by_sig, plot_width_by_sig

class ErrorTrainer:
    def __init__(self,
                 iterations,
                 train_data,
                 val_data,
                 loging_path="../runs",
                 num_classes=72,
                 model_path=None,
                 device="cuda:0",
                 data_folder="../data"):
        self.iterations = iterations  # Iteration refers to passes through all dataset
        self.num_classes = num_classes
        self.train_dataset = train_data
        self.val_dataset = val_data
        self.device = device
        self.model_path = model_path
        self.logger = ErrorFinderLogger(
            path=None,
            experiment_id=None)
        self.data_folder = data_folder

    def __loss(self,
               label,
               pred_lower,
               pred_upper):
        lagrange_base = float(self.loss_params["lagrange_base"])
        lagrange_high_error_sigs = float(self.loss_params["lagrange_high_error_sigs"])
        lagrange_pnorm = float(self.loss_params["lagrange_pnorm"])
        pnorm_order = int(self.loss_params["pnorm_order"] )
        lagrange_smalltozero = float(self.loss_params["lagrange_smalltozero"])
        pnorm_order = pnorm_order if pnorm_order%2 == 1 else pnorm_order + 1
        
        _EPS = 1e-6
        batch_size = float(pred_lower.shape[0])
        lower = label - pred_lower
        upper = pred_upper - label
        lower = nn.ReLU()(-lower)  # 0 if lower than label
        upper = nn.ReLU()(-upper)  # 0 if higher than label

        interval_length = ((pred_upper - pred_lower)**2)/batch_size
        
        # This is super sketchy
        lagrange_vector = torch.ones(72).to(self.device)*lagrange_base
        lagrange_vector[2] = lagrange_high_error_sigs
        lagrange_vector[4] = lagrange_high_error_sigs
        lagrange_vector[22] = lagrange_high_error_sigs
        lagrange_vector[43] = lagrange_high_error_sigs
        loss_by_mutation_signature =\
            interval_length +\
            lagrange_vector*(lower + upper)

        # p-norm by signature to avoid high errors
        loss_by_mutation = linalg.norm(lagrange_pnorm *\
            loss_by_mutation_signature, ord=pnorm_order, axis=1)
        loss = torch.mean(loss_by_mutation)

        # Send small to 0
        loss += lagrange_smalltozero*\
            torch.mean(torch.abs(pred_upper[label <= _EPS]))
        return loss

    def __meta_loss(self,
                    label,
                    pred_lower,
                    pred_upper):
        _EPS = 1e-6
        batch_size = float(pred_lower.shape[0])
        lower = label - pred_lower
        upper = pred_upper - label
        lower = nn.ReLU()(-lower)  # 0 if lower than label
        upper = nn.ReLU()(-upper)  # 0 if higher than label

        # Manage this case
        # assert(torch.isnan(pred_upper).any() == False)  # int len
        # assert(torch.isnan(pred_lower).any() == False)  # int len

        # Get interval length
        interval_length = torch.mean((pred_upper - pred_lower)**2)

        precision_by_sig = torch.mean((lower + upper < _EPS).to(torch.float32), dim=0)
        high_error_sigs = torch.sum(precision_by_sig<0.95)
        meta_loss = interval_length + high_error_sigs
        return meta_loss

    def objective(self,
                  batch_size,
                  lr,
                  num_neurons_pos,
                  num_neurons_neg,
                  num_hidden_layers_pos,
                  num_hidden_layers_neg,
                  loss_params,
                  plot=False):
        print(batch_size, lr, num_neurons_pos, num_hidden_layers_pos,
              num_neurons_neg, num_hidden_layers_neg, loss_params)
        self.loss_params = loss_params
        dataloader = DataLoader(dataset=self.train_dataset,
                                batch_size=int(batch_size),
                                shuffle=True)
        model = ErrorFinder(num_classes=self.num_classes,
                            num_hidden_layers_pos=int(num_hidden_layers_pos),
                            num_units_pos=int(num_neurons_pos),
                            num_hidden_layers_neg=int(num_hidden_layers_neg),
                            num_units_neg=int(num_neurons_neg))
        model.to(self.device)

        log_freq = 5

        optimizer = optim.Adam(model.parameters(),
                               lr=lr)
        meta_loss_vals = collections.deque(maxlen=20)
        max_found = -np.inf
        step = 0
        for _ in range(self.iterations):
            for _, train_label, train_weight_guess, num_mut, classification in tqdm(dataloader):
                optimizer.zero_grad()
                if "cuda" in str(self.device):
                    train_label = train_label.to(self.device)
                    train_weight_guess = train_weight_guess.to(self.device)
                    num_mut = num_mut.to(self.device)
                    classification = classification.to(self.device)
                train_pred_upper, train_pred_lower = model(weights=train_weight_guess,
                                                           num_mutations=num_mut,
                                                           classification=classification)

                # Compute loss
                train_loss = self.__loss(label=train_label,
                                         pred_lower=train_pred_lower,
                                         pred_upper=train_pred_upper)
                if torch.isnan(train_loss).any():
                    print("NANs appeared while training! Ending training now")
                    return -2.0
                train_loss.backward()
                optimizer.step()

                with torch.no_grad():
                    val_pred_upper, val_pred_lower = model(weights=self.val_dataset.prev_guess,
                                                           num_mutations=self.val_dataset.num_mut,
                                                           classification=self.val_dataset.classification)
                    val_loss = self.__loss(label=self.val_dataset.labels,
                                           pred_lower=val_pred_lower,
                                           pred_upper=val_pred_upper)
                    val_meta_loss = self.__meta_loss(label=self.val_dataset.labels,
                                                     pred_lower=val_pred_lower,
                                                     pred_upper=val_pred_upper)
                    meta_loss_vals.append(val_meta_loss.item())
                    max_found = max(max_found, -np.nanmean(meta_loss_vals))

                if plot and step % log_freq == 0:
                    pi_metrics_train = get_pi_metrics(train_label, train_pred_lower, train_pred_upper)
                    pi_metrics_val = get_pi_metrics(self.val_dataset.labels, val_pred_lower, val_pred_upper, collapse=False, dim=1)
                    self.logger.log(train_loss=train_loss,
                                    pi_metrics_train=pi_metrics_train,
                                    val_loss=val_loss,
                                    pi_metrics_val=pi_metrics_val,
                                    val_values_lower=val_pred_lower,
                                    val_values_upper=val_pred_upper,
                                    val_nummut=self.val_dataset.num_mut,
                                    step=step)
                del train_label, train_weight_guess, num_mut, classification, train_pred_upper, train_pred_lower, val_pred_upper, val_pred_lower
                # if self.model_path is not None and step % 500 == 0:
                #     save_model(model=model, directory=self.model_path)
                step += batch_size
        if self.model_path is not None:
            save_model(model=model, directory=self.model_path)

        with torch.no_grad():
            val_pred_upper, val_pred_lower = model(weights=self.val_dataset.prev_guess,
                                    num_mutations=self.val_dataset.num_mut,
                                    classification=self.val_dataset.classification)
            fig_error = plot_error_by_sig(label=self.val_dataset.labels.cpu(),
                                    pred_upper=val_pred_upper.cpu(),
                                    pred_lower=val_pred_lower.cpu(),
                                    sigs_names=list(pd.read_excel(self.data_folder + "/data.xlsx").columns)[1:])
            wandb.log({"Validation Error by Sig": wandb.Image(fig_error)})

            fig_width = plot_width_by_sig(pred_upper=val_pred_upper.cpu(),
                                        pred_lower=val_pred_lower.cpu(),
                                        sigs_names=list(pd.read_excel(self.data_folder + "/data.xlsx").columns)[1:])
            wandb.log({"Validation Width by Sig": wandb.Image(fig_width)})

        return max_found


def train_errorfinder(config, data_folder="data") -> float:
    from signaturesnet.utilities.io import read_data
    from modules.combined_finetuner import baseline_guess_to_combined_finetuner_guess

    # Select training device
    dev = "cuda" if config["device"] == "cuda" and torch.cuda.is_available() else "cpu"
    device = torch.device(dev)
    print("Using device:", device)

    # Set paths
    errorfinder_path = os.path.join(config["models_dir"], config["model_id"])
    classifier_path = os.path.join(config["models_dir"], config["classifier_id"])
    finetuner_low_path = os.path.join(config["models_dir"], config["finetuner_low_id"])
    finetuner_large_path = os.path.join(config["models_dir"], config["finetuner_large_id"])

    if config["enable_logging"]:
        run = wandb.init(project=config["wandb_project_id"],
                entity='sig-net',
                config=config,
                name=config["model_id"])

    # Load data
    train_real_low, val_real_low = read_data(experiment_id=config["data_id"],
                                            source="low",
                                            data_folder=data_folder,
                                            device=dev)
    train_real_large, val_real_large = read_data(experiment_id=config["data_id"],
                                                source="large",
                                                data_folder=data_folder,
                                                device=dev)
    # Join datasets
    train_data = train_real_low
    train_data.append(train_real_large)
    train_data.perm()

    val_data = val_real_low
    val_data.append(val_real_large)

    model = CombinedFinetuner(low_mum_mut_dir=finetuner_low_path,
                              large_mum_mut_dir=finetuner_large_path,
                              device=dev)
    classifier = read_model(classifier_path, device=dev)

    train_data = baseline_guess_to_combined_finetuner_guess(model=model,
                                                            classifier=classifier,
                                                            data=train_data)
    val_data = baseline_guess_to_combined_finetuner_guess(model=model,
                                                          classifier=classifier,
                                                          data=val_data)

    # train_data.to(device=device)  # We keep the train data in cpu to save memory
    # val_data.to(device=device)  # If still a problem we can make the val_data smaller
    # torch.cuda.empty_cache()

    trainer = ErrorTrainer(iterations=config["iterations"],  # Passes through all dataset
                           train_data=train_data,
                           val_data=val_data,
                           num_classes=config["num_classes"],
                           device=device,
                           model_path=errorfinder_path,
                           data_folder=data_folder)

    min_val = trainer.objective(batch_size=config["batch_size"],
                                lr=config["lr"],
                                num_neurons_pos=config["num_neurons_pos"],
                                num_neurons_neg=config["num_neurons_neg"],
                                num_hidden_layers_pos=config["num_hidden_layers_pos"],
                                num_hidden_layers_neg=config["num_hidden_layers_neg"],
                                loss_params=config["loss_params"],
                                plot=config["enable_logging"])
    if config["enable_logging"]:
        run.finish()
    return min_val


import torch
import wandb

class NumMutLogger:
    def __init__(self):
        pass

    def log(self,
            train_loss,
            train_classification_metrics,
            val_loss,
            val_classification_metrics,
            step):

        wandb.log({"train_loss": train_loss})
        wandb.log({"val_loss": val_loss})

        for metric in train_classification_metrics:
            wandb.log({"train_" + metric: train_classification_metrics[metric]})

        for metric in val_classification_metrics:
            wandb.log({"val_" + metric: val_classification_metrics[metric]})


import logging
import os
import pathlib
from sklearn import preprocessing
import sys
import yaml

import json
import pandas as pd
import torch

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from signaturesnet import DATA, TRAINED_MODELS
from signaturesnet.utilities.data_partitions import DataPartitions
from signaturesnet.utilities.generator_data import GeneratorData

def read_signatures(file,
                    mutation_type_order=os.path.join(DATA, "mutation_type_order.xlsx")):
    """
    File must contain first column with mutations types X[Y>Z]W and the rest of the columns must be the set of signatures
    """
    # Sort according to cosmic mutation types order
    signatures_data = sort_signatures(file,
                                      mutation_type_order=mutation_type_order)

    num_sigs = len(signatures_data.columns) - 1
    signatures = [torch.tensor(signatures_data.iloc[:, i]).type(torch.float32)
                  for i in range(1, num_sigs + 1)][:num_sigs]
    signatures = torch.stack(signatures).t()
    return signatures

def sort_signatures(file,
                    output_file=None,
                    mutation_type_order=os.path.join(DATA, "mutation_type_order.xlsx")):
    signatures_data = pd.read_excel(file)
    mutation_order = pd.read_excel(mutation_type_order)

    # Sort according to cosmic mutation types order
    signatures_data.rename(columns = {list(signatures_data)[0]:'Type'}, inplace=True)
    signatures_data = signatures_data.set_index('Type')
    signatures_data = signatures_data.reindex(index=mutation_order['Type'])
    signatures_data = signatures_data.reset_index()

    if output_file is not None:
        create_dir(output_file)
        signatures_data.to_csv(output_file, index=False)
    return signatures_data

def csv_to_pandas(file,
                  device="cpu",
                  header=None,
                  index_col=None,
                  type_df=None):
    df = pd.read_csv(file, header=header, index_col=index_col)
    df.index = df.index.map(lambda x: x.split("..")[-1])

    if type_df is not None:
        cancer_type_df = pd.read_csv(type_df, header=0)[["Cancer Types", "Sample Names"]]
        df = df.merge(cancer_type_df, left_index=True, right_on="Sample Names").set_index("Sample Names")
        le = preprocessing.LabelEncoder()
        le.fit(df["Cancer Types"])
        df['cancer_type'] = le.transform(df["Cancer Types"])
        df = df.drop("Cancer Types", axis=1)
    return df

def csv_to_tensor(file,
                  device="cpu",
                  header=None,
                  index_col=None,
                  type_df=None,):
    df = pd.read_csv(file, header=header, index_col=index_col)

    if type_df is not None:
        df.index = df.index.map(lambda x: x.split("..")[-1])
        cancer_type_df = pd.read_csv(type_df, header=0)[["Cancer Types", "Sample Names"]]
        df = df.merge(cancer_type_df, left_index=True, right_on="Sample Names").set_index("Sample Names")
        le = preprocessing.LabelEncoder()
        le.fit(df["Cancer Types"])
        df['cancer_type'] = le.transform(df["Cancer Types"])
        df = df.drop("Cancer Types", axis=1)

    input_tensor = torch.tensor(df.values, dtype=torch.float)
    assert(not torch.isnan(input_tensor).any())
    return input_tensor.float().to(device)

def tensor_to_csv(data_tensor, output_path):
    create_dir(output_path)
    df = data_tensor.detach().cpu().numpy()
    df = pd.DataFrame(df)
    df.to_csv(output_path, header=False, index=False) 

def read_data(device,
              experiment_id,
              source,
              data_folder=DATA,
              include_baseline=True,
              include_labels=True,
              n_points = None):
    """Read data from disk

    Args:
        device (string): Device to train on
        experiment_id (string): Full name of the experiment folder
        source (string): Type of generated data: random or realistic
        data_folder (str, optional): Relative path of data folder. Defaults to DATA.
    """
    # assert(source in ["random", "realistic", "perturbed"])
    path = os.path.join(data_folder, experiment_id)

    train_input = csv_to_tensor(path + "/train_%s_input.csv" % source, device)
    train_baseline = csv_to_tensor(path + "/train_%s_baseline.csv" % source, device) if include_baseline else None
    train_label = csv_to_tensor(path + "/train_%s_label.csv" % source, device) if include_labels else None
    train_data = DataPartitions(inputs=train_input,
                                prev_guess=train_baseline,
                                labels=train_label)
    train_data.perm
    if n_points is not None:
        train_data.inputs = train_data.inputs[:n_points,:]
        train_data.prev_guess = train_data.prev_guess[:n_points,:]
        train_data.labels = train_data.labels[:n_points,:]

    val_input = csv_to_tensor(path + "/val_%s_input.csv" % source, device)
    val_baseline = csv_to_tensor(path + "/val_%s_baseline.csv" % source, device) if include_baseline else None
    val_label = csv_to_tensor(path + "/val_%s_label.csv" % source, device) if include_labels else None
    
    val_data = DataPartitions(inputs=val_input,
                              prev_guess=val_baseline,
                              labels=val_label)

    return train_data, val_data

def read_data_classifier(device,
                         experiment_id,
                         data_folder=DATA):
    """Read data from disk

    Args:
        device (string): Device to train on
        experiment_id (string): Full name of the experiment folder
        source (string): Type of generated data: random or realistic
        data_folder (str, optional): Relative path of data folder. Defaults to "DATA".
    """
    path = os.path.join(data_folder, experiment_id)

    train_input = csv_to_tensor(path + "/train_input.csv", device)
    train_num_mut = csv_to_tensor(path + "/train_num_mut.csv", device).reshape((-1,1))
    train_label = csv_to_tensor(path + "/train_label.csv", device).reshape((-1,1))

    train_data = DataPartitions(inputs=train_input,
                                num_mut=train_num_mut,
                                labels=train_label)

    val_input = csv_to_tensor(path + "/val_input.csv", device)
    val_num_mut = csv_to_tensor(path + "/val_num_mut.csv", device).reshape((-1,1))
    val_label = csv_to_tensor(path + "/val_label.csv", device).reshape((-1,1))

    val_data = DataPartitions(inputs=val_input,
                              num_mut=val_num_mut,
                              labels=val_label)

    return train_data, val_data

def read_real_data(device, experiment_id, data_folder=DATA):
    """Read data from disk

    Args:
        device (string): Device to train on
        experiment_id (string): Full name of the experiment folder
        source (string): Type of generated data: random or realistic
        data_folder (str, optional): Relative path of data folder. Defaults to DATA.
    """
    path = os.path.join(data_folder, experiment_id)

    real_input = csv_to_tensor(path + "/real_data_input.csv", device)
    real_num_mut = csv_to_tensor(path + "/real_data_num_mut.csv", device)

    return real_input, real_num_mut


def read_data_generator(device,
                        data_id,
                        data_folder=DATA,
                        cosmic_version='v3',
                        type='real',
                        prop_train=0.9):
    '''
    type should be: 'real', 'perturbed' or 'augmented_real'.
    '''
    data_folder = data_folder + data_id
    if type == 'real':
        if cosmic_version == 'v3':
            real_data = csv_to_pandas(data_folder + "/sigprofiler_not_norm_PCAWG.csv",
                                    device=device,
                                    header=0,
                                    index_col=0,
                                    type_df=data_folder + "/PCAWG_sigProfiler_SBS_signatures_in_samples_v3.csv")
            
            num_ctypes = real_data['cancer_type'][-1]+1
            real_data = real_data.groupby('cancer_type').sample(frac=1, random_state=0)       #Shuffle samples inside the same cancer type
            # print(real_data.size(0)/num_ctypes*prop_train)
            real_data_train = real_data.groupby('cancer_type').head(int(round(real_data.shape[0]/num_ctypes*prop_train)))  #Take the first prop_train % of samples in each cancer type
            real_data_rest = pd.concat([real_data, real_data_train]).drop_duplicates(keep=False)                    # The rest is for validation and testing

            real_data_train = torch.tensor(real_data_train.values, dtype=torch.float)
            real_data_rest = torch.tensor(real_data_rest.values, dtype=torch.float)

            train_input, train_cancer_types = real_data_train[:, :-1], real_data_train[:, -1]
            train_input = train_input/torch.sum(train_input, axis=1).reshape(-1, 1)
            train_input = torch.cat([train_input, torch.zeros(train_input.size(0), 7).to(train_input)], dim=1)
            train_data = GeneratorData(inputs=train_input, cancer_types=train_cancer_types)

            val_input, val_cancer_types = real_data_rest[:, :-1], real_data_rest[:, -1]
            val_input = val_input/torch.sum(val_input, axis=1).reshape(-1, 1)
            val_input = torch.cat([val_input, torch.zeros(val_input.size(0), 7).to(val_input)], dim=1)
            val_data = GeneratorData(inputs=val_input, cancer_types=val_cancer_types)

        elif cosmic_version == 'v2':
            real_data = csv_to_tensor(data_folder + "/PCAWG_genome_deconstructSigs_v2.csv",
                                    device=device, header=0, index_col=0)
            real_data = real_data/torch.sum(real_data, axis=1).reshape(-1, 1)

            perm = torch.randperm(real_data.size(0))
            data = real_data[perm, :]

            train_input = data[:int(real_data.size(0)*0.95)]
            val_input = data[int(real_data.size(0)*0.95):]

            train_data = GeneratorData(inputs=train_input)
            val_data = GeneratorData(inputs=val_input)

        else:
            raise NotImplementedError

        
    else:
        train_input = csv_to_tensor(data_folder + "/train_%s_low_label.csv"%type,
                                    device=device, header=None, index_col=None)
        val_input = csv_to_tensor(data_folder + "/val_%s_low_label.csv"%type,
                                    device=device, header=None, index_col=None)
        train_data = GeneratorData(inputs=train_input[:,:-1])
        val_data = GeneratorData(inputs=val_input[:,:-1])

    train_data.to(device)
    val_data.to(device)
    return train_data, val_data

def read_methods_guesses(device, experiment_id, methods, data_folder=DATA):
    """Read one method guess from disk

    Args:
        device (string): Device to train on
        experiment_id (string): Full name of the experiment folder
        test_id (string): Full name of the test folder
        method (list): List of string with the methods to be analyzed
        data_folder (str, optional): Relative path of data folder. Defaults to DATA.
    """
    if experiment_id is not None:
        path = os.path.join(data_folder, experiment_id)

        methods_guesses = []
        for method in methods:
            methods_guesses.append(csv_to_tensor(
                path + "/other_methods/all_results/%s_guess.csv" % (method), device))

        label = csv_to_tensor(path + "/test_label.csv", device)

        return methods_guesses, label
    else:
        methods_guesses = []
        for method in methods:
            methods_guesses.append(csv_to_tensor(
                data_folder + "/%s_guess.csv" % (method), device))

        label = csv_to_tensor(DATA + "/datasets/test_label.csv", device)

        return methods_guesses, label


def read_test_data(device, experiment_id, test_id, data_folder=DATA):
    """Read one method guess from disk

    Args:
        device (string): Device to train on
        experiment_id (string): Full name of the experiment folder
        test_id (string): Full name of the test folder
        data_folder (str, optional): Relative path of data folder. Defaults to DATA.
    """
    path = os.path.join(data_folder, experiment_id, test_id)

    inputs = csv_to_tensor(path + "/%s_input.csv" % (test_id), device=device)
    label = csv_to_tensor(path + "/%s_label.csv" % (test_id), device=device)

    return inputs, label


def read_model(directory, device="cpu"):
    """Instantiate a pre-trained model from the stored vars
    The model is in cpu and in eval mode

    Args:
        directory (String): Folder containing state_dict and init_args.json of the model
    """
    from signaturesnet.models import Generator, Classifier, FineTunerLowNumMut, FineTunerLargeNumMut, ErrorFinder, NumMutNet

    # Load init_args
    init_args_file = os.path.join(directory, 'init_args.json')
    with open(init_args_file, 'r') as fp:
        init_args = json.load(fp)
    model_type = init_args["model_type"]
    init_args.pop("model_type")
    # print("Reading model of type:", model_type)
    assert(model_type is not None)  # Model type not saved!
    assert(model_type in ["Classifier", "FineTunerLowNumMut", "FineTunerLargeNumMut", "ErrorFinder", "Generator", "NumMutNet"])
    if "device" in init_args.keys():
        init_args["device"] = device
        
    # Instantiate model class
    if model_type == "Generator":
        model = Generator(**init_args)
    if model_type == "Classifier":
        model = Classifier(**init_args)
    elif model_type == "FineTunerLowNumMut":
        model = FineTunerLowNumMut(**init_args)
    elif model_type == "FineTunerLargeNumMut":
        model = FineTunerLargeNumMut(**init_args)
    elif model_type == "ErrorFinder":
        model = ErrorFinder(**init_args)
    elif model_type == "NumMutNet":
        model = NumMutNet(**init_args)
    
    # Load model weights
    state_dict_file = os.path.join(directory, "state_dict")
    try:
        state_dict = torch.load(f=state_dict_file,
                                map_location=torch.device(device))
    except:
        state_dict = torch.load(f=state_dict_file + ".zip",
                                map_location=torch.device(device))

    model.load_state_dict(state_dict)
    model.eval()
    model.to(device)
    return model

def save_model(model, directory):
    """Store a pytorch model. The arguments are splitted into 2 files:
    A init_args.json needed to instantiate the class, and the model state_dict

    Args:
        model (nn.Module): Model to save
        directory (String): Path where to save
    """
    pathlib.Path(directory).mkdir(parents=True, exist_ok=True)
    
    # Store init_args into a json file
    init_args_file = os.path.join(directory, 'init_args.json')
    with open(init_args_file, 'w') as fp:
        json.dump(model.init_args, fp)

    # Store state_dict
    state_dict_file = os.path.join(directory, "state_dict")
    torch.save(model.state_dict(), state_dict_file)


def update_dict(config, args):
    for arg in vars(args):
        if getattr(args, arg) is not None and arg in config:
            if len(getattr(args, arg)) == 1:
                config[arg] = getattr(args, arg)[0]
            else:
                config[arg] = getattr(args, arg)
    return config

def read_config(path):
    with open(path, 'r') as stream:
        data = yaml.safe_load(stream)
    return data["config"]

def create_dir(filepath):
    directory = os.path.dirname(filepath)
    pathlib.Path(directory).mkdir(parents=True, exist_ok=True)

def write_result(result, filepath):
    directory = os.path.dirname(filepath)
    pathlib.Path(directory).mkdir(parents=True, exist_ok=True)
    fout = open(filepath, 'w')
    fout.write(str(result))
    fout.close()

def write_final_outputs(weights,
                        lower_bound,
                        upper_bound,
                        classification,
                        sample_names, 
                        output_path,
                        name=''):
    pathlib.Path(output_path).mkdir(parents=True, exist_ok=True)
    sig_names = list(pd.read_excel(os.path.join(DATA, "data.xlsx")).columns)[1:]

    def write(values, path, sig_names, output_type = 'error'):
        logging.info("Writting results: %s"%path)
        try:
            df = pd.DataFrame(values.detach().numpy())
        except:
            df = pd.DataFrame(values)
        if output_type == 'weight':
            sig_names = sig_names + ['Unknown']
        if output_type == 'classification':
            sig_names = ['Classification']
        df.columns = sig_names
        df.index = sample_names
        df.to_csv(path, header=True, index=True)

    # Write results weight guesses
    write(values=weights, path=output_path + "/weight_guesses-%s.csv"%name, sig_names=sig_names, output_type='weight')
    write(values=lower_bound, path=output_path + "/lower_bound_guesses-%s.csv"%name, sig_names=sig_names)
    write(values=upper_bound, path=output_path + "/upper_bound_guesses-%s.csv"%name, sig_names=sig_names)
    write(values=classification, path=output_path + "/classification_guesses-%s.csv"%name, sig_names=sig_names, output_type='classification')


def write_David_outputs(weights, lower_bound, upper_bound, output_path):
    sig_names = list(pd.read_excel(os.path.join(DATA, "data.xlsx")).columns)[1:]
    
    # Write results weight guesses
    df = pd.DataFrame({'weight_guess': weights[0], 'upper_bound': upper_bound[0], 'lower_bound': lower_bound[0],})
    df.index = sig_names
    df.to_csv(output_path + "_guess.csv", header=True, index=True)




import torch
import torch.nn as nn
from torch.nn.modules.activation import Sigmoid


class FineTuner(nn.Module):
    def __init__(self,
                 num_classes=72,
                 num_hidden_layers=2,
                 num_units=400,
                 cutoff=0.01,
                 sigmoid_params=[[500, 1000], [5000, 2000], [10000, 5000]]):
        super(FineTuner, self).__init__()
        self.init_args = locals()
        self.init_args.pop("self")
        self.init_args.pop("__class__")
        self._cutoff = cutoff
        self._EPS = 1e-6
        self.sigmoid_params = sigmoid_params
        self.num_classes = num_classes

    def forward(self, *args, **kwargs):
        raise NotImplementedError

    def _apply_cutoff(self, comb, cutoff):
        mask = (comb > cutoff).type(torch.int).float()
        comb = comb*mask
        comb = torch.cat((comb, torch.ones_like(torch.sum(
        comb, axis=1).reshape((-1, 1)))-torch.sum(
        comb, axis=1).reshape((-1, 1))), axis=1)
        return comb

class FineTunerLowNumMut(FineTuner):

    def __init__(self,
                 num_classes=72,
                 num_hidden_layers=2,
                 num_units=400,
                 cutoff=0.01,
                 sigmoid_params=[5000, 2000]):
        super(FineTunerLowNumMut, self).__init__(num_classes=num_classes,
                                                 num_hidden_layers=num_hidden_layers,
                                                 num_units=num_units,
                                                 cutoff=0.01,
                                                 sigmoid_params=sigmoid_params)
        self.init_args["model_type"] = "FineTunerLowNumMut"

        # Num units of the mutations path
        num_units_branch_mut = 3
        num_units_joined_path = num_units + 3*num_units_branch_mut

        # Mutvec path
        self.layer_mutvec_1 = nn.Linear(96, num_units)
        self.layer_mutvec_2 = nn.Linear(num_units, num_units)

        # Nummut path
        self.layer_numut_low = nn.Linear(1, num_units_branch_mut)
        self.layer_numut_mid = nn.Linear(1, num_units_branch_mut)
        self.layer_numut_large = nn.Linear(1, num_units_branch_mut)
        self.layer_numut_joint = nn.Linear(
            3*num_units_branch_mut, 3*num_units_branch_mut)

        self.hidden_layers = nn.ModuleList(
            modules=[nn.Linear(num_units_joined_path, num_units_joined_path)
                     for _ in range(num_hidden_layers)])

        self.output_layer = nn.Linear(num_units_joined_path, num_classes)
        self.activation = nn.LeakyReLU(0.1)
        # self.layer_norm_1 = nn.LayerNorm(num_units_joined_path)
        # self.layer_norm_2 = nn.LayerNorm(num_units_joined_path)

        self.softmax = nn.Softmax(dim=1)
        self.dropout = nn.Dropout(p=0.1)
        # self.layer_norm = nn.LayerNorm(normalized_shape = num_units_joined_path)
        

    def forward(self,
                mutation_dist,
                num_mut,
                cutoff):
        # Input head
        mutation_dist = self.activation(self.layer_mutvec_1(mutation_dist))
        mutation_dist = self.activation(self.layer_mutvec_2(mutation_dist))

        # Number of mutations head
        num_mut_low = nn.Sigmoid()((num_mut - self.sigmoid_params[0][0]) / self.sigmoid_params[0][1])
        num_mut_mid = nn.Sigmoid()((num_mut - self.sigmoid_params[1][0]) / self.sigmoid_params[1][1])
        num_mut_large = nn.Sigmoid()((num_mut - self.sigmoid_params[2][0]) / self.sigmoid_params[2][1])
        num_mut_low = self.activation(self.layer_numut_low(num_mut_low))
        num_mut_mid = self.activation(self.layer_numut_mid(num_mut_mid))
        num_mut_large = self.activation(self.layer_numut_large(num_mut_large))
        num_mut = torch.cat([num_mut_low, num_mut_mid, num_mut_large], dim=1)
        num_mut = self.activation(self.layer_numut_joint(num_mut))


        # Concatenate
        comb = torch.cat([mutation_dist, num_mut], dim=1)
        # comb = self.layer_norm_1(comb)

        assert(not torch.isnan(comb).any())

        # comb = self.layer_norm(comb)
        # Apply shared layers
        for layer in self.hidden_layers:
            comb = self.activation(layer(self.dropout(comb)))

        # comb = self.layer_norm_2(comb)

        # Apply output layer
        comb = self.output_layer(comb)
        # comb += baseline_guess
        comb = self.softmax(comb)

        

        # If in eval mode, send small values to 0
        if not self.training:
            comb = self._apply_cutoff(comb, cutoff=cutoff)
        return comb


class FineTunerLargeNumMut(FineTuner):

    def __init__(self,
                 num_classes=72,
                 num_hidden_layers=2,
                 num_units=400,
                 cutoff=0.01,
                 sigmoid_params=[5000, 2000]):
        super(FineTunerLargeNumMut, self).__init__(num_classes=num_classes,
                                                 num_hidden_layers=num_hidden_layers,
                                                 num_units=num_units,
                                                 cutoff=0.01,
                                                 sigmoid_params=sigmoid_params)
        self.init_args["model_type"] = "FineTunerLargeNumMut"

        # Num units of the mutations path
        num_units_branch_mut = 10
        num_units_joined_path = 2*num_units + num_units_branch_mut

        self.layer1_1 = nn.Linear(num_classes, num_units)  # Baseline guess path
        # 96 = total number of possible muts
        self.layer1_2 = nn.Linear(96, num_units)  # Input path
        # Number of mutations path
        self.layer1_3 = nn.Linear(1, num_units_branch_mut)

        self.layer2_1 = nn.Linear(num_units, num_units)
        self.layer2_2 = nn.Linear(num_units, num_units)
        self.layer2_3 = nn.Linear(num_units_branch_mut, num_units_branch_mut)

        self.hidden_layers = nn.ModuleList(
            modules=[nn.Linear(num_units_joined_path, num_units_joined_path)
                     for _ in range(num_hidden_layers)])

        self.output_layer = nn.Linear(num_units_joined_path, num_classes)
        self.activation = nn.LeakyReLU(0.1)

        self.softmax = nn.Softmax(dim=1)
        self.dropout = nn.Dropout(p=0.1)

    def forward(self,
                mutation_dist,
                baseline_guess,
                num_mut,
                cutoff):
        # Input head
        mutation_dist = self.activation(self.layer1_2(mutation_dist))
        # mutation_dist = self.activation(self.layer2_2(mutation_dist))

        # Baseline head
        weights = self.activation(self.layer1_1(baseline_guess))
        # weights = self.activation(self.layer2_1(weights))

        # Number of mutations head
        num_mut = torch.log10(num_mut)/6
        num_mut = self.activation(self.layer1_3(num_mut))
        # num_mut = self.activation(self.layer2_3(num_mut))

        # Concatenate
        comb = torch.cat([mutation_dist, weights, num_mut], dim=1)

        # Apply shared layers
        for layer in self.hidden_layers:
            comb = self.activation(layer(self.dropout(comb)))

        # Apply output layer
        comb = self.output_layer(comb)
        # comb += baseline_guess
        comb = self.softmax(comb)

        # If in eval mode, send small values to 0
        if not self.training:
            comb = self._apply_cutoff(comb, cutoff=cutoff)
        return comb


def baseline_guess_to_finetuner_guess(trained_finetuner_dir, data):
    # Load finetuner and compute guess_1
    import gc
    from signaturesnet.utilities.io import read_model

    finetuner = read_model(directory=trained_finetuner_dir)
    finetuner.to("cpu")

    with torch.no_grad():
        data.prev_guess = finetuner(mutation_dist=data.inputs,
                                    baseline_guess=data.prev_guess,
                                    num_mut=data.num_mut,
                                    cutoff=0.01)
    del finetuner
    gc.collect()
    torch.cuda.empty_cache()
    return data



import torch


def small_to_unknown(sig_weights_batch, cutoff=0.01):
    """Send all values < cutoff to 0, adding their sum to a new element

    Args:
        sig_weights_batch (torch.tensor): Batch of signature weights to normalie
        cutoff (float): Value from which to send to zero
    """
    mask = (sig_weights_batch > cutoff).to(torch.int)
    sig_weights_batch = mask*sig_weights_batch
    result = torch.zeros((sig_weights_batch.shape[0], sig_weights_batch.shape[1] + 1)).to(sig_weights_batch)
    result[:, :-1] = sig_weights_batch
    result[:, -1] = (1 - torch.sum(sig_weights_batch, axis=1))
    return result


if __name__ == "__main__":
    example = torch.tensor([
        [0.1, 0.7, 0.2],
        [0.05, 0.1, 0.85],
        [0.5, 0.5, 0],
        [0.5, 0.5, 0],
    ])

    cutoff = 0.15

    example_with_unknown = small_to_unknown(example, cutoff)

    print(example)
    print(example_with_unknown)



import torch
import torch.nn as nn
import torch.nn.functional as F

class NumMutNet(nn.Module):
    def __init__(self,
                 hidden_dim,
                 n_layers = 3,
                 input_dim = 72,
                 output_dim = 9):
        self.init_args = locals()
        self.init_args.pop("self")
        self.init_args.pop("__class__")
        self.init_args["model_type"] = "NumMutNet"

        super(NumMutNet, self).__init__()

        self.input = nn.Linear(input_dim, hidden_dim)
        self.dropout = nn.Dropout(0.25)
        layers = []
        for i in range(n_layers):
            layers.append(nn.Linear(hidden_dim, hidden_dim))
        self.layers = nn.ModuleList(layers)
        self.output = nn.Linear(hidden_dim, output_dim)
        self.activation = nn.LeakyReLU(0.1)

    def forward(self, x):
        x = self.dropout(self.activation(self.input(x)))
        for layer in self.layers:
            x = self.dropout(self.activation(layer(x)))
        x = self.output(x)
        return x

    def get_nummuts(self, x):
        logits = self.forward(x)
        classes = torch.argmax(logits, dim=1)
        rand_ = torch.rand(classes.shape)
        classes_copy = classes.clone()
        classes = classes.to(torch.float)

        classes[classes_copy == 0] = rand_[classes_copy == 0] * (1.5 - 1) + 1
        classes[classes_copy == 1] = rand_[classes_copy == 1] * (2 - 1.5) + 1.5
        classes[classes_copy == 2] = rand_[classes_copy == 2] * (2.5 - 2) + 2
        classes[classes_copy == 3] = rand_[classes_copy == 3] * (3 - 2.5) + 2.5
        classes[classes_copy == 4] = rand_[classes_copy == 4] * (3.5 - 3) + 3
        classes[classes_copy == 5] = rand_[classes_copy == 5] * (4 - 3.5) + 3.5
        classes[classes_copy == 6] = rand_[classes_copy == 6] * (4.5 - 4) + 4
        classes[classes_copy == 7] = rand_[classes_copy == 7] * (5 - 4.5) + 4.5

        muts = torch.pow(10, classes).to(torch.float)
        return muts




import torch
import torch.nn as nn
from torch.nn.modules.activation import Sigmoid


class Classifier(nn.Module):

    def __init__(self,
                 num_hidden_layers=2,
                 num_units=400,
                 sigmoid_params=[5000, 1000]):
        self.init_args = locals()
        self.init_args.pop("self")
        self.init_args.pop("__class__")
        self.init_args["model_type"] = "Classifier"

        super(Classifier, self).__init__()
        
        self.sigmoid_params = sigmoid_params
        num_units_branch_mut = 10
        num_units_joined_path = num_units + num_units_branch_mut

        # Input path
        # 96 = total number of possible muts
        self.layer1_1 = nn.Linear(96, num_units)
        # Number of mutations path
        self.layer1_2 = nn.Linear(1, num_units_branch_mut)

        self.layer2_1 = nn.Linear(num_units, num_units)
        self.layer2_2 = nn.Linear(num_units_branch_mut, num_units_branch_mut)

        self.hidden_layers = nn.ModuleList(
            modules=[nn.Linear(num_units_joined_path, num_units_joined_path)
                     for _ in range(num_hidden_layers)])

        self.output_layer = nn.Linear(num_units_joined_path, 1)
        self.activation = nn.LeakyReLU(0.1)

        self.sigmoid = nn.Sigmoid()

    def forward(self,
                mutation_dist,
                num_mut):
        # Input head
        mutation_dist = self.activation(self.layer1_1(mutation_dist))
        mutation_dist = self.activation(self.layer2_1(mutation_dist))

        # Number of mutations head
        num_mut = nn.Sigmoid()(
            (num_mut-self.sigmoid_params[0])/self.sigmoid_params[1])
        num_mut = self.activation(self.layer1_2(num_mut))
        num_mut = self.activation(self.layer2_2(num_mut))

        # Concatenate
        comb = torch.cat([mutation_dist, num_mut], dim=1)

        # Apply shared layers
        for layer in self.hidden_layers:
            comb = self.activation(layer(comb))

        # Apply output layer
        comb = self.output_layer(comb)
        comb = self.sigmoid(comb)

        return comb



import logging
import numpy as np
import torch 

class ClassifiedFinetunerErrorfinder:

    def __init__(self,
                 classifier,
                 finetuner,
                 errorfinder,
                 classification_cutoff=0.5,
                 device="cpu"):
        """Instantiate a ClassifiedFinetuner

        Args:
            classifier (Classifier): Model to discriminate between realistic and random data
            realistic_finetuner (Finetuner or CombinedFinetuner): Model which improves baseline guess for random data
            random_finetuner (Finetuner or CombinedFinetuner): Model which improves baseline guess for random data
            classification_cutoff (float, optional): Cuttoff at which we decide something is realistic. Defaults to 0.5.
            device (str, optional): Device to use (cuda or cpu). Defaults to "cpu".
        """

        self.classification_cutoff = classification_cutoff
        self.device = device

        self.classifier = classifier
        self.finetuner = finetuner
        self.errorfinder = errorfinder

    def __separate_classification(self, classification, mutation_dist, baseline_guess, num_mut):
        ind = torch.tensor(range(classification.size()[0]))
        ind_order = torch.tensor(np.concatenate((ind[torch.flatten((classification <= self.classification_cutoff) | (torch.transpose(num_mut > 5e5,0,1)))], ind[torch.flatten((classification > self.classification_cutoff) & (torch.transpose(num_mut < 5e5,0,1)))]))).reshape(-1, 1).to(torch.float).to(self.device)
        
        input_batch_random = mutation_dist[torch.flatten((classification <= self.classification_cutoff) | (torch.transpose(num_mut > 5e5,0,1))), ]  # Random + samples with large num mut (we will only do NNLS for them)
        input_batch_realistic = mutation_dist[torch.flatten((classification > self.classification_cutoff) & (torch.transpose(num_mut < 5e5,0,1))), ]
        num_mut_realistic = num_mut[torch.flatten((classification > self.classification_cutoff) & (torch.transpose(num_mut < 5e5,0,1))), ]
        classification_realistic = classification[torch.flatten((classification > self.classification_cutoff) & (torch.transpose(num_mut < 5e5,0,1))), ]

        baseline_guess_random = baseline_guess[torch.flatten((classification <= self.classification_cutoff) | (torch.transpose(num_mut > 5e5,0,1))), ]
        baseline_guess_realistic = baseline_guess[torch.flatten((classification > self.classification_cutoff) & (torch.transpose(num_mut < 5e5,0,1))), ]

        baseline_guess_random = baseline_guess_random/torch.sum(baseline_guess_random, dim=1).reshape(-1,1)
        
        return input_batch_realistic, input_batch_random, baseline_guess_random, baseline_guess_realistic, num_mut_realistic, classification_realistic, ind_order

    def __join_and_sort(self, realistic, random, ind_order):
        joined = torch.cat((random, realistic), dim=0)
        joined = torch.cat((joined, ind_order), dim=1)
        joined = joined[joined[:, -1].sort()[1]]
        return joined[:, :-1]

    def _apply_cutoff(self, comb, cutoff):
        mask = (comb > cutoff).type(torch.int).float()
        comb = comb*mask
        comb = torch.cat((comb, torch.ones_like(torch.sum(
        comb, axis=1).reshape((-1, 1)))-torch.sum(
        comb, axis=1).reshape((-1, 1))), axis=1)
        return comb

    def __call__(self,
                 mutation_dist,
                 baseline_guess,
                 num_mut,
                 cutoff):

        logging.info("Detecting out-of-train-distribution points...")
        classification = self.classifier(mutation_dist=mutation_dist,
                                         num_mut=num_mut).view(-1)
        logging.info("Detecting out-of-train-distribution points... DONE")
        
        mutation_dist_realistic, mutation_dist_random, baseline_guess_random, baseline_guess_realistic, num_mut_realistic, classification_realistic, ind_order = self.__separate_classification(classification, mutation_dist, baseline_guess, num_mut)

        logging.info("Finetuning NNLS guesses...")
        finetuner_guess_realistic = self.finetuner(mutation_dist=mutation_dist_realistic,
                                         baseline_guess = baseline_guess_realistic,
                                         num_mut=num_mut_realistic,
                                         cutoff_0=cutoff)

        baseline_guess_random = self._apply_cutoff(baseline_guess_random, cutoff)

        finetuner_guess = self.__join_and_sort(finetuner_guess_realistic, baseline_guess_random, ind_order)
        logging.info("Finetuning NNLS guesses... DONE")

        logging.info("Estimating errorbars...")
        upper, lower = self.errorfinder(weights=finetuner_guess_realistic[:,:-1],
                                                num_mutations=num_mut_realistic,
                                                classification=classification_realistic.reshape(-1, 1))
        upper = self.__join_and_sort(upper, torch.full_like(baseline_guess_random[:,:-1], float('nan')), ind_order)
        lower = self.__join_and_sort(lower, torch.full_like(baseline_guess_random[:,:-1], float('nan')), ind_order)
        logging.info("Estimating errorbars... DONE")

        result = {"finetuner_guess": finetuner_guess,
                  "error_upper": upper,
                  "error_lower": lower,
                  "classification": classification}
        return result


import torch
import torch.nn as nn


class SignedErrorFinder(nn.Module):
    def __init__(self,
                 num_classes=72,
                 num_hidden_layers=2,
                 num_units=400,
                 sigmoid_params = [5000, 2000]):
        super(SignedErrorFinder, self).__init__()
        self.sigmoid_params = sigmoid_params
        # Num units of the weights path
        num_units_branch_1 = int(num_units*0.9)
        # Num units of the mutations path
        num_units_branch_2 = int((num_units - num_units_branch_1)/2)
        num_units_branch_3 = num_units - num_units_branch_1 - num_units_branch_2

        # Weights path
        self.layer1_1 = nn.Linear(num_classes, num_units_branch_1)
        nn.init.kaiming_uniform_(self.layer1_1.weight)
        self.layer2_1 = nn.Linear(num_units_branch_1, num_units_branch_1)
        nn.init.kaiming_uniform_(self.layer2_1.weight)

        # Mutations path
        self.layer1_2 = nn.Linear(1, num_units_branch_2)
        nn.init.kaiming_uniform_(self.layer1_2.weight)

        # Mutations path
        self.layer1_3 = nn.Linear(1, num_units_branch_3)
        nn.init.kaiming_uniform_(self.layer1_3.weight)

        # Combined layers
        self.hidden_layers = nn.ModuleList(
            modules=[nn.Linear(num_units, num_units)
                     for _ in range(num_hidden_layers)])
        self.output_layer = nn.Linear(num_units, num_classes)
        nn.init.kaiming_uniform_(self.output_layer.weight)
        for hidden_layer in self.hidden_layers:
            nn.init.kaiming_uniform_(hidden_layer.weight)

        # Activation function
        self.activation = nn.LeakyReLU(0.1)

    def __clamp(self, x, slope=1e-2):
        return nn.LeakyReLU(slope)(1 - nn.LeakyReLU(slope)(1 - x))

    def forward(self, weights, num_mutations, classification):
        # Baseline head
        weights = self.activation(self.layer1_1(weights))
        weights = self.activation(self.layer2_1(weights))

        # Mutations head
        num_mutations = torch.log10(num_mutations)/6
        # num_mutations = nn.Sigmoid()((num_mutations-self.sigmoid_params[0])/self.sigmoid_params[1])
        num_mutations = self.activation(self.layer1_2(num_mutations))

        # Baseline head
        classification = self.activation(self.layer1_3(classification))

        # Concatenate
        comb = torch.cat([weights, num_mutations, classification], dim=1)

        # Apply shared layers
        for layer in self.hidden_layers:
            comb = self.activation(layer(comb))

        # Apply output layer
        # res = self.__clamp(self.output_layer(comb))
        res = self.output_layer(comb)
        return res


class ErrorFinder(nn.Module):

    def __init__(self,
                 num_classes=72,
                 num_hidden_layers_pos=2,
                 num_hidden_layers_neg=2,
                 num_units_pos=400,
                 num_units_neg=400):
        self.init_args = locals()
        self.init_args.pop("self")
        self.init_args.pop("__class__")
        self.init_args["model_type"] = "ErrorFinder"
        
        super(ErrorFinder, self).__init__()

        self.positive_path = SignedErrorFinder(num_classes=num_classes,
                                               num_hidden_layers=num_hidden_layers_pos,
                                               num_units=num_units_pos)

        self.negative_path = SignedErrorFinder(num_classes=num_classes,
                                               num_hidden_layers=num_hidden_layers_neg,
                                               num_units=num_units_neg)

    def forward(self, weights, num_mutations, classification):
        pred_upper = self.positive_path(weights, num_mutations, classification)
        pred_lower = self.negative_path(weights, num_mutations, classification)

        # If in eval mode, correct interval issues
        if not self.training:
            pred_upper[pred_upper < weights] = weights[pred_upper < weights]
            pred_lower[pred_lower > weights] = weights[pred_lower > weights]
            pred_lower[pred_lower < 0] = 0
            pred_upper[pred_upper > 1] = 1

        return pred_upper, pred_lower



from argparse import ArgumentParser
import os
import sys

import torch
import wandb

from signaturesnet.trainers import train_errorfinder
from signaturesnet.utilities.io import update_dict, read_config, write_result

DEFAULT_CONFIG_FILE = ["configs/errorfinder/errorfinder.yaml"]
torch.manual_seed(123)

if __name__ == "__main__":
    # Parse command-line arguments
    parser = ArgumentParser()
    parser.add_argument(
        '--config_file', action='store', nargs=1, type=str, required=False, default=DEFAULT_CONFIG_FILE,
        help=f'Path to yaml file containing all needed parameters.\nThey can be overwritten by command-line arguments'
    )

    parser.add_argument(
        '--model_id', action='store', nargs=1, type=str, required=False,
        help=f'Unique id given to the trained model.'
    )
    parser.add_argument(
        '--source', action='store', nargs=1, type=str, required=False,
        help=f'Data source to use in the training.'
    )

    # Train args
    parser.add_argument(
        '--batch_size', action='store', nargs=1, type=int, required=False,
        help='Training batch size.'
    )
    parser.add_argument(
        '--lr', action='store', nargs=1, type=float, required=False,
        help='Learning rate.'
    )
    parser.add_argument(
        '--sigmoid_params', action='store', nargs=1, type=list, required=False,
        help='Sigmoid parameters for normalization.'
    )
    # Model args
    parser.add_argument(
        '--num_neurons_pos', action='store', nargs=1, type=int, required=False,
        help='Neurons of the positive error network.'
    )
    parser.add_argument(
        '--num_hidden_layers_pos', action='store', nargs=1, type=int, required=False,
        help='Layers of the positive error network.'
    )
    parser.add_argument(
        '--num_neurons_neg', action='store', nargs=1, type=int, required=False,
        help='Neurons of the negative error network.'
    )
    parser.add_argument(
        '--num_hidden_layers_neg', action='store', nargs=1, type=int, required=False,
        help='Layers of the negative error network.'
    )

    # Loss args
    parser.add_argument(
        '--lagrange_base', action='store', nargs=1, type=float, required=False,
        help='Weight of misclassification per base signatures.'
    )
    parser.add_argument(
        '--lagrange_high_error_sigs', action='store', nargs=1, type=float, required=False,
        help='Weight of misclassification per high error signatures.'
    )
    parser.add_argument(
        '--lagrange_pnorm', action='store', nargs=1, type=float, required=False,
        help='Weight of big errors in the loss.'
    )
    parser.add_argument(
        '--lagrange_smalltozero', action='store', nargs=1, type=float, required=False,
        help='Weight of not sending small values to zero in the loss.'
    )
    parser.add_argument(
        '--pnorm_order', action='store', nargs=1, type=int, required=False,
        help='Norm used in the loss.'
    )
    _args = parser.parse_args()

    # Read & config
    config = read_config(path=getattr(_args, "config_file")[0])
    config = update_dict(config=config, args=_args)
    config["loss_params"] = update_dict(config=config["loss_params"], args=_args)
    
    print("Using config:", config)
    score = train_errorfinder(config=config)
    # write_result(score, "../tmp/errorfinder_score_%s_%s.txt"%(config["source"],config["model_id"]))


import os
import sys

import torch
import numpy as np

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from signaturesnet import DATA
from models import Baseline
from signaturesnet.utilities.data_generator import DataGenerator
from signaturesnet.utilities.data_partitions import DataPartitions
from signaturesnet.utilities.io import csv_to_pandas, read_signatures, csv_to_tensor
from signaturesnet.utilities.oversampler import CancerTypeOverSampler


def read_data_nummutnet(path, device):
    # TODO read input and labels
    train_input = csv_to_tensor(file=path + "/train_input.csv", device=device)
    train_label = csv_to_tensor(file=path + "/train_label.csv", device=device).to(torch.long)
    val_input = csv_to_tensor(file=path + "/val_input.csv", device=device)
    val_label = csv_to_tensor(file=path + "/val_label.csv", device=device).to(torch.long)

    train_data = DataPartitions(inputs=train_input,
                                labels=train_label,
                                extract_nummut=False)
    val_data = DataPartitions(inputs=val_input,
                              labels=val_label,
                              extract_nummut=False)
    return train_data, val_data

def generate_realistic_nummut_data(
                                    n_samples,
                                    data_id="real_data",
                                    data_folder=DATA,
                                    split="train",
                                    large_or_low="large",
                                    device="cuda",):
    # Read real weight combinations
    labels_filepath = os.path.join(data_folder, data_id, "PCAWG_sigProfiler_SBS_signatures_in_samples_v3.csv")
    df = csv_to_pandas(labels_filepath, header=0, index_col=0)
    weights = torch.from_numpy(df.iloc[:, 2:].values)
    # Add zeros for last 7 missing signatures
    weights = torch.cat([weights, torch.zeros(weights.size(0), 7).to(weights)], dim=1)
    nummuts = torch.sum(weights, axis=1).to(torch.float32).unsqueeze(1)

    # Create inputs associated to the labels:
    signatures = read_signatures(
        os.path.join(DATA, "data.xlsx"),
        mutation_type_order=os.path.join(DATA, "mutation_type_order.xlsx"))
    data_generator = DataGenerator(signatures=signatures,
                                   seed=None,
                                   normalize=True,
                                   shuffle=True)
    inputs, labels = data_generator.make_input(labels=weights,
                                                split=split,
                                                large_low=large_or_low,
                                                nummuts=nummuts)
    inputs = inputs[:n_samples]
    labels = labels[:n_samples]

    # Run Baseline
    baseline = Baseline(signatures)
    train_baseline = baseline.get_weights_batch(inputs, n_workers=4)
    
    # Split into train and validation
    data = DataPartitions(inputs=inputs.float(),
                          prev_guess=train_baseline.float(),
                          labels=labels.float())
    data.to(device)
    return data


def read_data_final_finetuner(device, data_id, data_folder=DATA, network_type="low"):
    '''
    Read all real data, oversample and generate samples with different numbers of mutations
    to train the final finetuner.
    '''
    data_folder = data_folder + data_id
    real_data = csv_to_pandas(data_folder + "/sigprofiler_not_norm_PCAWG.csv",
                                    device=device, header=0, index_col=0,
                                    type_df=data_folder + "/PCAWG_sigProfiler_SBS_signatures_in_samples_v3.csv")
    ctypes = torch.tensor(real_data.values[:,-1], dtype=torch.float)
    real_data = torch.tensor(real_data.values[:,:-1], dtype=torch.float)
    real_data = torch.cat([real_data, torch.zeros(real_data.size(0), 7).to(real_data)], dim=1)
    oversampler = CancerTypeOverSampler(real_data, ctypes)
    oversampled_weights = oversampler.get_N_oversampled_set(N_samples=1)

    # Create inputs associated to the labels:
    signatures = read_signatures(
        DATA + "data.xlsx",
        mutation_type_order=DATA + "mutation_type_order.xlsx")
    data_generator = DataGenerator(signatures=signatures,
                                   seed=None,
                                   shuffle=True)

    train_input, train_label = data_generator.make_input(oversampled_weights, "train", network_type, normalize=True)
    
    # Run Baseline
    sf = Baseline(signatures)
    train_baseline = sf.get_weights_batch(train_input, n_workers=2)
    
    train_data = DataPartitions(inputs=train_input.float().to(device),
                                prev_guess=train_baseline.float().to(device),
                                labels=train_label.float().to(device))
    val_data = train_data
    return train_data, val_data



import torch
from torch import nn

from signaturesnet.utilities.data_generator import DataGenerator

class Generator(nn.Module):
    
    def __init__(self,
                 input_size=72,
                 num_hidden_layers=2,
                 latent_dim=100,
                 device="cuda") -> None:
        self.init_args = locals()
        self.init_args.pop("self")
        self.init_args.pop("__class__")
        self.init_args["model_type"] = "Generator"
        super(Generator, self).__init__()

        self.latent_dim = latent_dim
        encoder_layers = []
        for i in range(num_hidden_layers):
            in_features = int(input_size - (input_size-latent_dim)*i/num_hidden_layers)
            out_features = int(input_size - (input_size-latent_dim)*(i+1)/num_hidden_layers)
            print("in_features:", in_features, "out_features:", out_features)
            layer = nn.Linear(in_features, out_features)
            # layernorm = torch.nn.LayerNorm(in_features)
            # encoder_layers.append(layernorm)
            encoder_layers.append(layer)
        self.encoder_layers = nn.ModuleList(modules=encoder_layers)
        self.mean_layer = nn.Linear(latent_dim, latent_dim)
        self.var_layer = nn.Linear(latent_dim, latent_dim)

        decoder_layers = []
        for i in reversed(range(num_hidden_layers+1)):
            in_features = int(input_size - (input_size-latent_dim)*(i+1)/(num_hidden_layers + 1))
            out_features = int(input_size - (input_size-latent_dim)*i/(num_hidden_layers + 1))
            layer = nn.Linear(in_features, out_features)            
            # layernorm = torch.nn.LayerNorm(in_features)
            # decoder_layers.append(layernorm)
            decoder_layers.append(layer)

        self.decoder_layers = nn.ModuleList(modules=decoder_layers)
        self.activation = nn.LeakyReLU(0.1)
        self.relu = nn.ReLU()
        

        self.Normal = torch.distributions.Normal(0, 1)
        if device == "cuda":
            print("Sending generator to cuda")
            self.Normal.loc = self.Normal.loc.cuda()
            self.Normal.scale = self.Normal.scale.cuda()
        self.device = device


    def encode(self, x):
        for layer in self.encoder_layers:
            x = self.activation(layer(x))
        z_mu = self.mean_layer(x)
        z_var = torch.exp(self.var_layer(x))
        # z_var = torch.ones_like(z_mu)
        # z_var = torch.ones_like(z_mu)*0.7
        # z_var = torch.ones_like(z_mu)*0.0001
        return z_mu, z_var

    def decode(self, x):
        for layer in self.decoder_layers[:-1]:
            x = self.activation(layer(x))
        x = self.relu(self.decoder_layers[-1](x))
        x = x/x.sum(dim=1).reshape(-1,1)
        return x

    def forward(self, x, noise=True):
        z_mu, z_var = self.encode(x)
        if noise:
            # z = z_mu + z_var*self.Normal.sample(z_mu.shape)
            z = torch.randn(size = (z_mu.size(0), z_mu.size(1)), device=self.device)
            z = z_mu + z_var*z
        else:
            z = z_mu
        x = self.decode(z)
        return x, z_mu, z_var

    def generate_weights(self, n_samples:int, std = 1.0, cutoff = 0.01):
        shape = tuple((n_samples, self.latent_dim))
        z = self.Normal.sample(shape)*std
        labels = self.decode(z)
        if cutoff > 0:
            labels[labels < cutoff] = 0
            labels = labels/labels.sum(dim=1).reshape(-1,1)
        return labels

    def generate_weights_and_samples(self,
                                     n_samples,
                                     signatures,
                                     realistic_number_of_mutations=False,
                                     nummutnet=None,
                                     std = 1.0,
                                     cutoff = 0.01):
        """Generate a dataset of realistic signature weights and their sampled mutation vectors.

        Args:
            n_samples (int): Number of samples to generate.
            signatures (torch.tensor): Signature matrix.
            realistic_number_of_mutations (bool, optional): Wether to use the nummut model to generate
                realistic number of mutations for each sample. Defaults to False.
            nummutnet (NumMutNet, optional): Model to estimate the number of mutations for a given signature weight combination. Defaults to None.
            std (float, optional): Standard deviation of latent variables. Defaults to 1.0.
            cutoff (float, optional): Minimum weight assigned t signature. Defaults to 0.01.
        
        Returns:
            inputs (mutational vector)
            labels (including an appended column with the number of mutations used)
        """
        labels = self.generate_weights(n_samples, std, cutoff)
        data_generator = DataGenerator(signatures)
        
        if realistic_number_of_mutations:
            assert nummutnet is not None  # Must provide a mutnet model!
            num_mutations = nummutnet.get_nummuts(labels)
        else:
            num_mutations = np.random.randint(1, 5e3, size=n_samples)

        return data_generator.make_input(labels,
                                        split=None,
                                        large_low=None,
                                        augmentation=1,
                                        nummuts=num_mutations)
    






import collections
import os
import sys

import numpy as np
import pandas as pd
from pandas.core.algorithms import mode
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from tqdm import tqdm
import wandb

from signaturesnet import DATA
from signaturesnet.utilities.metrics import get_jensen_shannon, get_kl_divergence
from signaturesnet.utilities.io import save_model
from signaturesnet.utilities.generator_data import GeneratorData
from signaturesnet.utilities.oversampler import OverSampler, CancerTypeOverSampler
from signaturesnet.models import Generator
from signaturesnet.loggers.generator_logger import GeneratorLogger

class GeneratorTrainer:
    def __init__(self,
                 iterations,
                 train_data,
                 val_data,
                 signatures,
                 lagrange_param=1.0,
                 loging_path="../runs",
                 num_classes=72,
                 log_freq=100,
                 model_path=None,  # File where to save model learned weights None to not save
                 device=torch.device("cuda:0")):
        self.iterations = iterations  # Now iteration refers to passes through all dataset
        self.num_classes = num_classes
        self.device = device
        self.log_freq = log_freq
        self.lagrange_param = lagrange_param
        self.model_path = model_path
        self.train_dataset = train_data
        self.val_dataset = val_data

        self.logger = GeneratorLogger(
            train_inputs=train_data.inputs,
            val_inputs=val_data.inputs,
            signatures=signatures,
            device=device)

        os = CancerTypeOverSampler(self.train_dataset.inputs, self.train_dataset.cancer_types)
        # os = OverSampler(self.train_dataset.inputs)
        self.train_dataset.inputs = os.get_oversampled_set()

    def __loss(self, input, pred, z_mu, z_std):
        kl_div = (0.5*(z_std.pow(2) + z_mu.pow(2) - 2*torch.log(z_std) - 1).sum(dim=1)).mean(dim=0)
        reconstruction = nn.MSELoss()(input, pred)
        # reconstruction = get_jensen_shannon(predicted_label=pred, true_label=input)
        return reconstruction + self.adapted_lagrange_param*kl_div

    def objective(self,
                  batch_size,
                  lr_encoder,
                  lr_decoder,
                  num_hidden_layers,
                  latent_dim,
                  plot=False):

        print(batch_size, lr_encoder, lr_decoder,
              num_hidden_layers, latent_dim)

        dataloader = DataLoader(dataset=self.train_dataset,
                                batch_size=int(batch_size),
                                shuffle=True)
        model = Generator(input_size=int(self.num_classes),
                          num_hidden_layers=int(num_hidden_layers),
                          latent_dim=int(latent_dim),
                          device=self.device.type)
        model.to(self.device)

        # wandb.watch(model, log_freq=100)

        # optimizer = optim.Adam(model.parameters(), lr=lr, weight_decay=1e-5)
        optimizer = optim.Adam([
            {'params': model.encoder_layers.parameters(), 'lr': lr_encoder},
            {'params': model.decoder_layers.parameters(), 'lr': lr_decoder}
        ])

        # l_vals = collections.deque(maxlen=50)
        # max_found = -np.inf
        step = 0
        # total_steps = 1000*len(self.train_dataset)
        total_steps = self.iterations*len(self.train_dataset)
        # self.batch_size_factor = batch_size/len(self.train_dataset)
        self.batch_size_factor = 1.
        train_DQ99R = None
        for iteration in range(self.iterations):
            for train_input in tqdm(dataloader):
                model.train()  # NOTE: Very important! Otherwise we zero the gradient
                optimizer.zero_grad()
                train_pred, train_mean, train_std = model(train_input)
                # self.adapted_lagrange_param = self.lagrange_param
                if step < total_steps*0.8:
                    self.adapted_lagrange_param = self.lagrange_param * \
                        float(total_steps - step)/float(total_steps)
                else:
                    self.adapted_lagrange_param = self.lagrange_param * \
                        float(total_steps - total_steps*0.8)/float(total_steps)
                train_loss = self.__loss(input=train_input,
                                         pred=train_pred,
                                         z_mu=train_mean,
                                         z_std=train_std)

                train_loss.backward()
                optimizer.step()

                model.eval()
                with torch.no_grad():
                    val_pred, val_mean, val_std = model(
                        self.val_dataset.inputs)
                    val_loss = self.__loss(input=self.val_dataset.inputs,
                                           pred=val_pred,
                                           z_mu=val_mean,
                                           z_std=val_std)
                    # l_vals.append(val_loss.item())
                    # max_found = max(max_found, -np.nanmean(l_vals))

                if plot and step % self.log_freq == 0:
                    current_train_DQ99R = self.logger.log(train_loss=train_loss,
                                                        train_prediction=train_pred,
                                                        train_label=train_input,
                                                        val_loss=val_loss,
                                                        val_prediction=val_pred,
                                                        val_label=self.val_dataset.inputs,
                                                        train_mu=train_mean,
                                                        train_sigma=train_std,
                                                        val_mu=val_mean,
                                                        val_sigma=val_std,
                                                        step=step,
                                                        model=model)
                    train_DQ99R = current_train_DQ99R if current_train_DQ99R is not None else train_DQ99R

                if self.model_path is not None and step % 500 == 0:
                    save_model(model=model, directory=self.model_path)
                step += 1
        if self.model_path is not None:
            save_model(model=model, directory=self.model_path)
        
        # Return last mse and KL obtained in validation
        return train_DQ99R

def log_results(config, train_DQ99R, out_csv_path):
    model_results = pd.DataFrame({"batch_size": [config["batch_size"]],
                                  "lr_encoder": [config["lr_encoder"]],
                                  "lr_decoder": [config["lr_decoder"]],
                                  "num_hidden_layers": [config["num_hidden_layers"]],
                                  "latent_dim": [config["latent_dim"]],
                                  "lagrange_param": [config["lagrange_param"]],
                                  "train_DQ99R": [train_DQ99R]})
    model_results.to_csv(out_csv_path,
                         header=False, index=False, mode="a")

def train_generator(config, data_folder=DATA + "/") -> float:
    """Train a classification model and get the validation score

    Args:
        config (dict): Including all the needed args
        to load data, and train the model 
    """
    from signaturesnet.utilities.io import read_data_generator, sort_signatures

    dev = "cuda" if config["device"] == "cuda" and torch.cuda.is_available(
    ) else "cpu"
    print("Using device:", dev)

    if config["enable_logging"]:
        run = wandb.init(project=config["wandb_project_id"],
                   entity='sig-net',
                   config=config,
                   name=config["model_id"])

    train_data, val_data = read_data_generator(device=dev,
                                               data_id=config['data_id'],
                                               cosmic_version=config['cosmic_version'],
                                               data_folder=data_folder,
                                               type=config['type'])

    signatures = sort_signatures(file=data_folder + "data.xlsx",
                                 mutation_type_order=data_folder + "mutation_type_order.xlsx")

    trainer = GeneratorTrainer(iterations=config["iterations"],  # Passes through all dataset
                               train_data=train_data,
                               val_data=val_data,
                               signatures=signatures,
                               lagrange_param=config["lagrange_param"],
                               num_classes=config["num_classes"],
                               device=torch.device(dev),
                               model_path=os.path.join(config["models_dir"], config["model_id"]))

    train_DQ99R = trainer.objective(batch_size=config["batch_size"],
                                    lr_encoder=config["lr_encoder"],
                                    lr_decoder=config["lr_decoder"],
                                    num_hidden_layers=config["num_hidden_layers"],
                                    latent_dim=config["latent_dim"],
                                    plot=config["enable_logging"])

    wandb.log({"train_DQ99R_score": train_DQ99R})

    if config["enable_logging"]:
        run.finish()
    return train_DQ99R



import torch
from torch.utils.data import Dataset

class GeneratorData(Dataset):
    def __init__(self, inputs=None, cancer_types=None):
        self.inputs = inputs
        self.cancer_types = cancer_types

    def __len__(self):
        return self.inputs.shape[0]

    def __getitem__(self, i):
        return self.inputs[i]
    
    def to(self, device):
        self.inputs = self.inputs.to(device)
        self.cancer_types = self.cancer_types.to(device)



from concurrent.futures import ProcessPoolExecutor
import copy
import sys

import pandas as pd
import scipy
from scipy.optimize import minimize
from scipy.optimize import nnls
import torch
from tqdm import tqdm

from signaturesnet import DATA
from signaturesnet.utilities.metrics import get_jensen_shannon
from signaturesnet.utilities.io import create_dir, read_signatures


class Baseline:

    def __init__(self, signatures):
        self.signatures = signatures.cpu().detach().numpy()
        self.__weight_len = self.signatures.shape[1]

    def get_weights(self, normalized_mutations):
        h, rnorm = nnls(self.signatures, normalized_mutations,
                        maxiter=5*self.__weight_len)
        return torch.from_numpy(h).float()

    def get_weights_batch(self, input_batch, n_workers=8):
        result = []
        input_as_list = input_batch.tolist()
        with ProcessPoolExecutor(max_workers=n_workers) as executor:
            for r in executor.map(self.get_weights, input_as_list):
                result.append(copy.deepcopy(r))
                del r
        guessed_labels = torch.stack(result)
        return guessed_labels


def create_baseline_dataset(input_file, output_file, signatures_path, which_baseline="nnls"):
    signatures = read_signatures(signatures_path)

    if which_baseline == "nnls":
        sf = Baseline(signatures)
    else:
        raise NotImplementedError("Other optimization methods have been deprecated")

    input_data = torch.tensor(pd.read_csv(
        input_file, header=None).values, dtype=torch.float)
    sol = sf.get_weights_batch(input_data)
    sol = sol.detach().numpy()

    df = pd.DataFrame(sol)
    create_dir(output_file)
    df.to_csv(output_file, header=False, index=False)
    print("Baseline Done!")


if __name__ == "__main__":

    cosmic_version = str(sys.argv[1])

    if cosmic_version == 'v3':
        signatures_file = DATA + "data.xlsx"
        experiment_id = "exp_not_norm"
    elif cosmic_version == 'v2':
        signatures_file = DATA + "data_v2.xlsx"
        experiment_id = "exp_generator_v2"
    else:
        raise NotImplementedError

    in_files = [
        # generator
        "/%s/train_generator_low_input.csv"%experiment_id,
        "/%s/train_generator_large_input.csv"%experiment_id,
        "/%s/val_generator_low_input.csv"%experiment_id,
        "/%s/val_generator_large_input.csv"%experiment_id,
        "/%s/train_perturbed_low_input.csv"%experiment_id,
        "/%s/train_perturbed_large_input.csv"%experiment_id,
        "/%s/val_perturbed_low_input.csv"%experiment_id,
        "/%s/val_perturbed_large_input.csv"%experiment_id,

        # tests
        "/%s/test_generator_input.csv"%experiment_id,
        "/%s/test_perturbed_input.csv"%experiment_id,
    ]    

    for in_file in in_files:
        print("Computing baseline of:", in_file)
        try:
            in_file = data_folder + in_file
            out_file = in_file.replace("input", "baseline")
            create_baseline_dataset(input_file=in_file,
                                    output_file=out_file,
                                    signatures_path=signatures_file,
                                    which_baseline="nnls")
        except Exception as e:
            print(e)


import torch
import wandb

class ErrorFinderLogger:
    def __init__(self, path, experiment_id):
        pass

    def log(self,
            train_loss,
            pi_metrics_train,
            val_loss,
            pi_metrics_val,
            val_values_lower,
            val_values_upper,
            val_nummut,
            step):

        wandb.log({"train_loss": train_loss})
        wandb.log({"val_loss": val_loss})

        for key in pi_metrics_train.keys():
            wandb.log({"train_%s"%key: pi_metrics_train[key].item()})
        
        for key in pi_metrics_val.keys():
            wandb.log({"val_%s"%key: torch.mean(pi_metrics_val[key]).item()})


from .classifier_trainer import ClassifierTrainer, train_classifier
from .finetuner_trainer import FinetunerTrainer, train_finetuner
from .error_trainer import ErrorTrainer, train_errorfinder
from .generator_trainer import GeneratorTrainer, train_generator
from .nummut_trainer import NumMutTrainer, train_nummutnet


from .baseline import Baseline
from .classifier import Classifier
from .errorfinder import ErrorFinder
from .finetuner import FineTunerLowNumMut, FineTunerLargeNumMut
from .generator import Generator
from .nummut import NumMutNet


import torch
from torch.utils.data import Dataset


class DataPartitions(Dataset):
    def __init__(self, inputs=None, labels=None, prev_guess=None, num_mut=None, extract_nummut=True):
        """Encapsulates input, label and previous step guess of a dataset.
        
        Args:
            inputs (Torch.Tensor, optional): Input to current ANN. Defaults to None.
            labels (torch.Tensor, optional): Labels of current inputs. Defaults to None.
            prev_guess (Guess of the previous model, optional): If training finetuner, 
                this is the baseline output, if training the errorfinder this is the
                finetuner output. Defaults to None.
        """
        self.inputs = inputs
        self.labels = labels
        self.prev_guess = prev_guess
        self.classification = None
        self.num_mut = num_mut if num_mut is not None else None
        if extract_nummut and labels is not None and num_mut is None:
            self.num_mut = labels[:, -1].reshape(-1,1)
            self.labels = labels[:, :-1]

    def append(self, other):
        def cat(a, b):
            if a is not None and b is not None:
                return torch.cat([a, b], dim=0) 
        self.inputs = cat(self.inputs, other.inputs)
        self.labels = cat(self.labels, other.labels)
        self.prev_guess = cat(self.prev_guess, other.prev_guess)
        self.classification = cat(self.classification, other.classification)
        self.num_mut = cat(self.num_mut, other.num_mut)

    def perm(self):
        indexes = torch.randperm(self.inputs.shape[0])
        def do_perm(a):
            if a is not None:
                return a[indexes, ...]
        self.inputs = do_perm(self.inputs)
        self.labels = do_perm(self.labels)
        self.prev_guess = do_perm(self.prev_guess)
        self.classification = do_perm(self.classification)
        self.num_mut = do_perm(self.num_mut)

    def to(self, device):
        device = torch.device(device)
        self.inputs = self.inputs.to(device)
        self.classification = self.classification.to(device) if self.classification is not None else self.labels
        self.labels = self.labels.to(device) if self.labels is not None else self.labels
        self.prev_guess = self.prev_guess.to(device) if self.prev_guess is not None else self.prev_guess
        self.num_mut = self.num_mut.to(device) if self.num_mut is not None else self.num_mut

    def __len__(self):
        return self.inputs.shape[0]

    def __getitem__(self, i):
        labels_i = torch.empty(0) if self.labels is None else self.labels[i]
        prev_guess_i = torch.empty(0) if self.prev_guess is None else self.prev_guess[i]
        num_mut_i = torch.empty(0) if self.num_mut is None else self.num_mut[i]
        classification_i = torch.empty(0) if self.classification is None else self.classification[i]
        return self.inputs[i], labels_i, prev_guess_i, num_mut_i, classification_i



from setuptools import setup, find_packages

# Read the content of your README file
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='signaturesnet',
      version='0.0.5',
      description="Package to manipulate mutational processes.",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/weghornlab/SigNet',
      packages=find_packages(),
      install_requires=[
            'torch',
            'scipy',
            'numpy',
            'matplotlib',
            'pandas',
            'seaborn',
            'scikit_optimize',
            'tqdm',
            'pyparsing',
            # 'gaussian_process',
            'PyYAML',
            'scikit_learn',
            'openpyxl',
            'tensorboard',
            'wandb',
      ])



import numpy as np
import matplotlib.pyplot as plt
import torch
from torch.utils.tensorboard import SummaryWriter
import wandb

from signaturesnet.utilities.metrics import get_MSE, get_kl_divergence, sets_distances, get_distances_metrics
from signaturesnet.utilities.plotting import get_correlation_matrix


class GeneratorLogger:
    def __init__(self, train_inputs, val_inputs, signatures, device):
        self.counter = 0
        self.plot_freq = 10
        self.metrics = {
            "mse": get_MSE,
        }
        self.train_inputs = train_inputs
        self.presence = torch.mean(train_inputs, dim=0)
        self.val_inputs = val_inputs
        self.signatures = signatures
        self.device = device

    def log(self,
            train_loss,
            train_prediction,
            train_label,
            train_mu,
            train_sigma, 
            val_loss,
            val_prediction,
            val_label,
            val_mu,
            val_sigma,
            model,
            step):

        wandb.log({"train_loss": train_loss})
        wandb.log({"val_loss": val_loss})

        for metric_name in self.metrics.keys():
            metric = self.metrics[metric_name]
            wandb.log({"train_" + metric_name: metric(train_prediction, train_label).item()})
            wandb.log({"val_" + metric_name: metric(val_prediction, val_label).item()})

        # val_mse = get_MSE(val_prediction, val_label).item()

        # Between-examples variance
        train_variance = torch.mean(torch.var(train_prediction, dim=0)).item()
        val_variance = torch.mean(torch.var(val_prediction, dim=0)).item()
        wandb.log({"train_pred_variance": train_variance})
        wandb.log({"val_pred_variance": val_variance})

        # Within-examples variance
        train_variance = torch.mean(torch.var(train_mu, dim=1)).item()
        val_variance = torch.mean(torch.var(val_mu, dim=1)).item()
        wandb.log({"train_mu_variance": train_variance})
        wandb.log({"val_mu_variance": val_variance})
        
        def KL(mu, sigma):
            return (0.5*(sigma + mu**2 - torch.log(sigma) - 1).sum(dim=1)).mean(dim=0)
        
        val_KL = KL(val_mu, val_sigma).item()
        wandb.log({"train_KL": KL(train_mu, train_sigma).item()})
        wandb.log({"val_KL": val_KL})

        wandb.log({"train_mu": torch.mean(train_mu)})
        wandb.log({"val_mu": torch.mean(val_mu)})

        wandb.log({"train_sigma": torch.mean(train_sigma)})
        wandb.log({"val_sigma": torch.mean(val_sigma)})

        train_DQ99R = None
        if self.counter % self.plot_freq == 0:
            n = self.train_inputs.size(0) + self.train_inputs.size(1)
            noise = torch.randn((n, model.latent_dim), device=self.device)
            with torch.no_grad():
                generated = model.decode(noise).detach()
            train_real_dists, train_fake_dists = sets_distances(
                real=self.train_inputs,
                fake=generated[:self.train_inputs.size(0)])
            val_real_dists, val_fake_dists = sets_distances(
                real=self.val_inputs,
                fake=generated[self.train_inputs.size(0):])
            train_real_metrics = get_distances_metrics(train_real_dists)
            train_fake_metrics = get_distances_metrics(train_fake_dists)
            val_real_metrics = get_distances_metrics(val_real_dists)
            val_fake_metrics = get_distances_metrics(val_fake_dists)
            train_DQ99R = train_real_metrics["quantiles"][-1]
            wandb.log({"train_DQ99R" : train_real_metrics["quantiles"][-1],
                       "train_DQ99G" : train_fake_metrics["quantiles"][-1]})
            wandb.log({"val_DQ99R" : val_real_metrics["quantiles"][-1],
                       "val_DQ99G" : val_fake_metrics["quantiles"][-1]})
            plt.switch_backend('agg')
            if self.counter % 50*self.plot_freq == 0:
                fig, _ = get_correlation_matrix(generated, self.signatures.columns[1:])
                wandb.log({"correlation": wandb.Image(fig)})
                plt.cla()
                plt.clf()
                plt.close()

                gen_presence = torch.mean(generated, dim=0)
                presence_errors = (gen_presence - self.presence).detach().cpu().numpy()
                # fig2 = plt.figure()
                plt.bar(np.array(list(range(presence_errors.shape[0]))), presence_errors)
                # plt.show()
                wandb.log({"error_by_signature": plt})
            
        self.counter += 1
        return train_DQ99R


import numpy as np
import pandas as pd
import torch

from signaturesnet import DATA

def normalize_data(input_file, opportunities_name_or_path):
    if opportunities_name_or_path == 'genome':
        return input_file
    else:
        genome_abundances = create_opportunities(DATA + '/real_data/3mer_WG_hg37.txt')
        genome_abundances_tensor = torch.Tensor(genome_abundances)

        if opportunities_name_or_path == 'exome':
            abundances = create_opportunities(DATA + '/real_data/abundances_trinucleotides.txt')
        elif opportunities_name_or_path == 'genome_hg38':
            abundances = create_opportunities(DATA + '/real_data/3mer_WG_hg38.txt')
        else:
            abundances = create_opportunities(opportunities_name_or_path)

        abundances_tensor = torch.Tensor(abundances)
        abundances_tensor = (abundances_tensor/genome_abundances_tensor)/torch.sum(abundances_tensor/genome_abundances_tensor)
        return torch.div(input_file, abundances_tensor)



def create_opportunities(abundances_path):
    # Create opportunities
    order = ['ACA','ACC','ACG','ACT','CCA','CCC','CCG','CCT','GCA','GCC','GCG','GCT','TCA','TCC','TCG','TCT',
             'ACA','ACC','ACG','ACT','CCA','CCC','CCG','CCT','GCA','GCC','GCG','GCT','TCA','TCC','TCG','TCT',
             'ACA','ACC','ACG','ACT','CCA','CCC','CCG','CCT','GCA','GCC','GCG','GCT','TCA','TCC','TCG','TCT',
             'ATA','ATC','ATG','ATT','CTA','CTC','CTG','CTT','GTA','GTC','GTG','GTT','TTA','TTC','TTG','TTT',
             'ATA','ATC','ATG','ATT','CTA','CTC','CTG','CTT','GTA','GTC','GTG','GTT','TTA','TTC','TTG','TTT',
             'ATA','ATC','ATG','ATT','CTA','CTC','CTG','CTT','GTA','GTC','GTG','GTT','TTA','TTC','TTG','TTT']
    
    opp_file = open(abundances_path, 'r')
    opp_file_lines = opp_file.readlines()
    opp_dic = {}
    for line in opp_file_lines:
        try:
            opp_dic[line.split("\t")[0]] = float(line.strip("\n").split("\t")[1])
        except:
            opp_dic[line.split(" ")[0]] = float(line.strip("\n").split(" ")[1])
    opp = np.zeros(96)
    for i in range(len(order)):
        tri = order[i]
        opp[i] = opp_dic[tri]
        b1 = tri[0]
        b2 = tri[1]
        b3 = tri[2]
        complement_base = complement(b3) + complement(b2) + complement(b1)
        opp[i] = opp[i] + opp_dic[complement_base]
    return opp



def complement(base):
    if base == "A":
        return "T"
    if base == "C":
        return "G"
    if base == "G":
        return "C"
    if base == "T":
        return "A"



import os

from torch.utils.tensorboard import SummaryWriter
import wandb

from signaturesnet.utilities.metrics import get_MSE,\
    get_negative_cosine_similarity,\
    get_cross_entropy2,\
    get_kl_divergence,\
    get_jensen_shannon,\
    get_wasserstein_distance



class FinetunerLogger:
    def __init__(self):
        self.metrics = {
            # "mse": get_MSE,
            # "cos": get_negative_cosine_similarity,
            # "cross_ent": get_cross_entropy2,
            "KL": get_kl_divergence,
            # "JS": get_jensen_shannon,
            # "W": get_wasserstein_distance,
        }

    def log(self,
            train_loss,
            train_prediction,
            train_label,
            train_classification_metrics,
            val_loss,
            val_prediction,
            val_label,
            val_classification_metrics,
            step):

        wandb.log({"train_loss": train_loss})
        wandb.log({"val_loss": val_loss})

        for metric_name in self.metrics.keys():
            metric = self.metrics[metric_name]
            wandb.log({"train_" + metric_name: metric(train_prediction, train_label).item()})
            wandb.log({"val_" + metric_name: metric(val_prediction, val_label).item()})

        for metric_name in train_classification_metrics.keys():
            try:
                wandb.log({"train_" + metric_name: train_classification_metrics[metric_name].item()})
            except:
                pass

        for metric_name in val_classification_metrics.keys():
            try:
                wandb.log({"val_" + metric_name: val_classification_metrics[metric_name].item()})
            except:
                pass



import os

import numpy as np
import torch
import torch.optim as optim
from torch.utils.data import DataLoader
from torchmetrics import Accuracy
from tqdm import tqdm
from signaturesnet.utilities.plotting import plot_confusion_matrix
import wandb

from signaturesnet.models import NumMutNet
from signaturesnet.loggers.nummut_logger import NumMutLogger
from signaturesnet.utilities.io import save_model
from signaturesnet import DATA, TRAINED_MODELS
from signaturesnet.utilities.temporal_io import read_data_nummutnet


class NumMutTrainer:
    def __init__(self,
                 train_data,
                 val_data,
                 iterations,
                 log_freq=100,
                 log_path="../runs",
                 model_path=None,
                 device=torch.device("cuda:0"),):
        self.train_dataset = train_data
        self.val_dataset = val_data
        self.iterations = iterations
        self.log_freq = log_freq
        self.log_path = log_path
        self.model_path = model_path
        self.device = device
        self.criterion = torch.nn.CrossEntropyLoss()
        self.logger = NumMutLogger()

    def objective(self,
                  batch_size,
                  lr,
                  num_hidden_layers,
                  num_units,
                  plot=False):

        dataloader = DataLoader(dataset=self.train_dataset,
                                batch_size=int(batch_size),
                                shuffle=True)
        model = NumMutNet(n_layers=int(num_hidden_layers),
                          hidden_dim=int(num_units))
        model.to(self.device)
        optimizer = optim.Adam(model.parameters(), lr=lr, weight_decay=1e-4)
        

        step = 0
        l_vals, max_found = [], -np.inf
        for _ in range(self.iterations):
            for train_input, train_label, _, _, _ in tqdm(dataloader):
                model.train()  # NOTE: Very important! Otherwise we zero the gradient
                optimizer.zero_grad()   
                train_prediction = model(train_input)

                train_loss = self.criterion(input=train_prediction,
                                            target=train_label.squeeze(),)

                train_loss.backward()
                optimizer.step()

                model.eval()
                with torch.no_grad():
                    one_hot_label = torch.nn.functional.one_hot(train_label, num_classes=train_prediction.shape[-1]).squeeze()
                    accuracy_train = Accuracy().to(self.device)(preds=train_prediction,
                                                target=one_hot_label)
                    val_prediction = model(self.val_dataset.inputs)
                    val_loss = self.criterion(input=val_prediction,
                                              target=self.val_dataset.labels.squeeze(),)
                    l_vals.append(val_loss.item())
                    max_found = max(max_found, -np.nanmean(l_vals))

                if plot and step % self.log_freq == 0:
                    one_hot_label = torch.nn.functional.one_hot(self.val_dataset.labels.squeeze(),
                                                                num_classes=train_prediction.shape[-1]).squeeze()
                    accuracy_val = Accuracy().to(self.device)(preds=val_prediction,
                                              target=one_hot_label)

                    self.logger.log(train_loss=train_loss,
                                    train_classification_metrics={"accuracy": accuracy_train},
                                    val_loss=val_loss,
                                    val_classification_metrics={"accuracy": accuracy_val},
                                    step=step)

                # if self.model_path is not None and step % 1000 == 0:
                #     save_model(model=model, directory=self.model_path + '_it' + str(step))
                step += 1
        save_model(model=model, directory=self.model_path)

        with torch.no_grad():
            val_prediction = model(self.val_dataset.inputs)
            val_prediction = torch.argmax(val_prediction, dim=1)
            label_val = self.val_dataset.labels.squeeze()
            labels = torch.unique(torch.cat((val_prediction, label_val))).tolist()
            conf_mat = plot_confusion_matrix(label_val, val_prediction, labels)
            wandb.log({"Confusion Matrix": wandb.Image(conf_mat)})

        return max_found

def train_nummutnet(config,
                    models_folder=TRAINED_MODELS,
                    data_folder=DATA):
    train_data, val_data = read_data_nummutnet(path=os.path.join(data_folder, "datasets/num_muts/"),
                                               device=config["device"],)
    
    if config["enable_logging"]:
        run = wandb.init(project=config["wandb_project_id"],
                         entity='signet_2',
                         config=config,
                         name=config["model_name"])


    trainer = NumMutTrainer(train_data=train_data,
                            val_data=val_data,
                            iterations=config["iterations"],
                            model_path=os.path.join(models_folder, config["model_name"]),
                            device=config["device"],)

    min_val = trainer.objective(batch_size=config["batch_size"],
                                lr=config["lr"],
                                num_hidden_layers=config["num_hidden_layers"],
                                num_units=config["num_neurons"],
                                plot=True)

    if config["enable_logging"]:
        wandb.log({"validation_score": min_val})
        run.finish()

    return min_val


import collections
import os

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from tqdm import tqdm
import wandb

from signaturesnet.loggers import ClassifierLogger
from signaturesnet.models import Classifier
from signaturesnet.utilities.data_partitions import DataPartitions
from signaturesnet.utilities.io import save_model

class ClassifierTrainer:
    def __init__(self,
                 iterations,
                 train_data,
                 val_data,
                 sigmoid_params = [5000, 2000],
                 loging_path="../runs",
                 num_classes=72,
                 log_freq=100,
                 model_path=None,  # File where to save model learned weights None to not save
                 cutoff=0.5,
                 device=torch.device("cuda:0")):
        self.iterations = iterations  # Now iteration refers to passes through all dataset
        self.num_classes = num_classes
        self.sigmoid_params = sigmoid_params
        self.device = device
        self.log_freq = log_freq
        self.model_path = model_path
        self.train_dataset = train_data
        self.val_dataset = val_data
        self.cutoff = cutoff
        self.logger = ClassifierLogger()

    def __loss(self, prediction, label):
        return nn.BCELoss()(prediction, label)
         
    def objective(self,
                  batch_size,
                  lr,
                  num_hidden_layers,
                  num_neurons,
                  run=None):

        print(batch_size, lr, num_hidden_layers, num_neurons)

        dataloader = DataLoader(dataset=self.train_dataset,
                                batch_size=int(batch_size),
                                shuffle=True)
        model = Classifier(num_hidden_layers=int(num_hidden_layers),
                          num_units=int(num_neurons),
                          sigmoid_params=self.sigmoid_params)
        model.to(self.device)

        optimizer = optim.Adam(model.parameters(), lr=lr)

        l_vals = collections.deque(maxlen=50)
        max_found = -np.inf
        step = 0
        for iteration in range(self.iterations):
            for train_input, train_label, _, num_mut, _ in tqdm(dataloader):
                model.train()  # NOTE: Very important! Otherwise we zero the gradient
                optimizer.zero_grad()                
                train_prediction = model(train_input, num_mut)
                train_loss = self.__loss(prediction=train_prediction,
                                         label=train_label)

                train_loss.backward()
                optimizer.step()

                model.eval()
                with torch.no_grad():
                    val_prediction = model(
                        self.val_dataset.inputs, self.val_dataset.num_mut)
                    val_loss = self.__loss(prediction=val_prediction,
                                        label=self.val_dataset.labels)
                    l_vals.append(val_loss.item())
                    max_found = max(max_found, -np.nanmean(l_vals))

                if run and step % self.log_freq == 0:
                    self.logger.log(train_loss=train_loss,
                                    train_prediction=(train_prediction > self.cutoff).type(torch.int64),
                                    train_label=train_label.type(torch.int64),
                                    val_loss=val_loss,
                                    val_prediction=(val_prediction > self.cutoff).type(torch.int64),
                                    val_label=self.val_dataset.labels.type(torch.int64),
                                    step=step)

                if self.model_path is not None and step % 500 == 0:
                    save_model(model=model, directory=self.model_path)
                step += 1
        if self.model_path is not None:
            save_model(model=model, directory=self.model_path)
        
        if run is not None:
            run.finish()
        return max_found


def train_classifier(config) -> float:
    """Train a classification model and get the validation score

    Args:
        config (dict): Including all the needed args
        to load data, and train the model 
    """
    from signaturesnet.utilities.io import read_data_classifier

    dev = "cuda" if config["device"] == "cuda" and torch.cuda.is_available() else "cpu"
    print("Using device:", dev)

    run = None
    if config["enable_logging"]:
        run = wandb.init(project=config["wandb_project_id"],
                        entity='sig-net',
                        config=config,
                        name=config["model_id"])

    train_data, val_data = read_data_classifier(device=dev,
                                                experiment_id=config["data_id"])

    trainer = ClassifierTrainer(iterations=config["iterations"],  # Passes through all dataset
                               train_data=train_data,
                               val_data=val_data,
                               sigmoid_params=config["sigmoid_params"],
                               device=torch.device(dev),
                               model_path=os.path.join(config["models_dir"], config["model_id"]))

    min_val = trainer.objective(batch_size=config["batch_size"],
                                lr=config["lr"],
                                num_hidden_layers=config["num_hidden_layers"],
                                num_neurons=config["num_neurons"],
                                run=run)

    return min_val


import os
import sys

import numpy as np
import pandas as pd
import torch

from signaturesnet.utilities.io import read_model

class CombinedFinetuner:
    def __init__(self,
                 low_mum_mut_dir,
                 large_mum_mut_dir,
                 cuttoff = 1e3,
                 device="cpu"):

        # Instantiate finetuner 1 and read params
        self.finetuner_low = read_model(low_mum_mut_dir, device=device)
        self.finetuner_large = read_model(large_mum_mut_dir, device=device)
        self.cutoff = cuttoff
        self.device = device

    def __join_and_sort(self, low, large, ind_order):
        joined = torch.cat((low, large), dim=0)
        joined = torch.cat((joined, ind_order), dim=1)
        joined = joined[joined[:, -1].sort()[1]]
        return joined[:, :-1]
    
    def __call__(self,
                 mutation_dist,
                 baseline_guess,
                 num_mut,
                 cutoff_0):
        """Get weights of each signature in lexicographic wrt 1-mer
        """
        num_mut = num_mut.view(-1)
        ind = torch.tensor(range(mutation_dist.size()[0]))
        ind_order = torch.tensor(np.concatenate((ind[num_mut <= self.cutoff], ind[num_mut > self.cutoff]))).reshape(-1, 1).to(torch.float).to(self.device)
        
        input_batch_low = mutation_dist[num_mut <= self.cutoff, ]
        input_batch_large = mutation_dist[num_mut > self.cutoff, ]

        baseline_guess_low = baseline_guess[num_mut <= self.cutoff, ]
        baseline_guess_large = baseline_guess[num_mut > self.cutoff, ]
        
        num_mut_low = num_mut[num_mut <= self.cutoff, ].reshape(-1, 1)
        num_mut_large = num_mut[num_mut > self.cutoff, ].reshape(-1, 1)
        
        with torch.no_grad():
            guess_low = self.finetuner_low(
                input_batch_low, num_mut_low, cutoff_0)

            guess_large = self.finetuner_large(
                input_batch_large, baseline_guess_large, num_mut_large, cutoff_0)
            
            finetuner_guess = self.__join_and_sort(low=guess_low,
                                                   large=guess_large,
                                                   ind_order=ind_order)
        return finetuner_guess


def baseline_guess_to_combined_finetuner_guess(model, classifier, data):
    # Load finetuner and compute guess_1 only when the sample is classified as realistic
    import gc
    with torch.no_grad():
        data.classification = classifier(mutation_dist=data.inputs,
                                         num_mut=data.num_mut)
        data.inputs = data.inputs[data.classification.view(-1) > 0.5, ]
        data.num_mut = data.num_mut[data.classification > 0.5, ].reshape(-1,1)
        data.prev_guess = model(mutation_dist=data.inputs,
                                baseline_guess=data.prev_guess,
                                num_mut=data.num_mut,
                                cutoff_0=0.01)
        data.prev_guess = data.prev_guess[:,:-1]
        data.classification = data.classification[data.classification > 0.5, ].reshape(-1,1)
    del model
    gc.collect()
    torch.cuda.empty_cache()
    return data


import os
import sys

from argparse import ArgumentParser
import torch
import pandas as pd
import wandb

from signaturesnet.trainers import train_generator
from signaturesnet.utilities.io import update_dict, read_config, write_result

DEFAULT_CONFIG_FILE = ["configs/generator/generator.yaml"]

if __name__ == "__main__":
    # Parse command-line arguments
    parser = ArgumentParser()
    parser.add_argument(
        '--config_file', action='store', nargs=1, type=str, required=False, default=DEFAULT_CONFIG_FILE,
        help=f'Path to yaml file containing all needed parameters.\nThey can be overwritten by command-line arguments'
    )

    parser.add_argument(
        '--model_id', action='store', nargs=1, type=str, required=False,
        help=f'Unique id given to the trained model.'
    )
    parser.add_argument(
        '--type', action='store', nargs=1, type=str, required=False,
        help=f'Type of the data that we are going to use.'
    )
    # Train args
    parser.add_argument(
        '--lagrange_param', action='store', nargs=1, type=float, required=False,
        help='Lagrange Parameter'
    )
    parser.add_argument(
        '--batch_size', action='store', nargs=1, type=int, required=False,
        help='Training batch size.'
    )
    parser.add_argument(
        '--lr_encoder', action='store', nargs=1, type=float, required=False,
        help='Learning rate of the encoder.'
    )
    parser.add_argument(
        '--lr_decoder', action='store', nargs=1, type=float, required=False,
        help='Learning rate of the decoder.'
    )
    # Model args
    parser.add_argument(
        '--latent_dim', action='store', nargs=1, type=int, required=False,
        help='Neurons of the positive error network.'
    )
    parser.add_argument(
        '--num_hidden_layers', action='store', nargs=1, type=int, required=False,
        help='Layers of the positive error network.'
    )

    _args = parser.parse_args()

    # Read & config
    config = read_config(path=getattr(_args, "config_file")[0])
    config = update_dict(config=config, args=_args)
    print("Using config:", config)
    train_DQ99R = train_generator(config=config)
    model_results = pd.DataFrame({"batch_size": [config["batch_size"]],
                                  "lr_encoder": [config["lr_encoder"]],
                                  "lr_decoder": [config["lr_decoder"]],
                                  "num_hidden_layers": [config["num_hidden_layers"]],
                                  "latent_dim": [config["latent_dim"]],
                                  "lagrange_param": [config["lagrange_param"]],
                                  "adapted_lagrange_param": [""],
                                  "batch_size_factor": [""],
                                  "val_mse": [val_mse],
                                  "val_KL": [val_KL],
                                  "val_loss": [""]})
    model_results.to_csv("../tmp/generator_models.csv",
                         header=False, index=False, mode="a")



import os
import pandas as pd
from pathlib import Path
import pysam
from collections import Counter
import genomepy

from signaturesnet import DATA


def complement(base):
    if base == "A":
        return "T"
    if base == "C":
        return "G"
    if base == "G":
        return "C"
    if base == "T":
        return "A"

# https://www.biostars.org/p/334253/
def VCF_to_counts(vcf_path, reference_genome):

    isFile = os.path.isfile(vcf_path)

    if isFile == True:
        list_of_files = [vcf_path]
    else:
        list_of_files = os.listdir(vcf_path)
        list_of_files = [vcf_path + '/' + file for file in list_of_files]
    # open fasta file
    if reference_genome == None:
        print("ERROR: You should provide a name or path to the reference genome of your data!")
        exit()
    else:
        try:
            genome = pysam.FastaFile(reference_genome)
        except:
            try:
                genome = pysam.FastaFile(DATA+'/genomes/'+reference_genome+'/'+reference_genome+'.fa')
            except:
                genomepy.install_genome(reference_genome,provider='UCSC',genomes_dir=DATA+'/genomes')
                genome = pysam.FastaFile(DATA+'/genomes/'+reference_genome+'/'+reference_genome+'.fa')
    mutation_type_order=os.path.join(DATA, "mutation_type_order.xlsx")
    mutation_order = pd.read_excel(mutation_type_order)
    final_df = pd.DataFrame(columns=mutation_order['Type'])
    for file in list_of_files:
        # open vcf file
        save = pysam.set_verbosity(0)
        vcf = pysam.VariantFile(file)
        pysam.set_verbosity(save)
        file_name = Path(file).stem
        # define by how many bases the variant should be flanked
        flank = 1
        # iterate over each variant
        list_of_mutations = []
        record_count = 0
        for record in vcf:
            # extract sequence
            #
            # The start position is calculated by subtract the number of bases
            # given by 'flank' from the variant position. The position in the vcf file
            # is 1-based. pysam's fetch() expected 0-base coordinate. That's why we
            # need to subtract on more base.
            #
            # The end position is calculated by adding the number of bases
            # given by 'flank' to the variant position. We also need to add the length
            # of the REF value and subtract again 1 due to the 0-based/1-based thing.
            # 
            # We need to check that 'chr' is contained in the chromosome, otherwise it doesn't work.
            if 'chr' not in record.chrom:
                chr = 'chr' + record.chrom
            else:
                chr = record.chrom
            # Now we have the complete sequence like this:
            # [number of bases given by flank]+REF+[number of bases given by flank]
            seq = genome.fetch(chr, record.pos-1-flank, record.pos-1+len(record.ref)+flank)
            # print out tab seperated columns:
            # CRHOM, POS, REF, ALT, flanking sequencing with variant given in the format '[REF/ALT]'
            mutation = '{}[{}>{}]{}'.format(seq[:flank].upper(), record.ref.upper(), record.alts[0].upper(), seq[flank+len(record.ref):].upper())
            list_of_mutations.append(mutation)
        # Now we count the number of times each mutation type is present in the list
        counts = Counter(list_of_mutations)
        # Filter and sort based on the mutation categories. 
        muts_sorted = []
        for mutation_type in mutation_order['Type']:
            mutation_type_rev = complement(mutation_type[6])+'['+complement(mutation_type[2])+'>'+complement(mutation_type[4])+']'+complement(mutation_type[0])
            muts_sorted.append(counts[mutation_type]+counts[mutation_type_rev])
        file_dict = {mutation_order.loc[i, 'Type']: [muts_sorted[i]] for i in range(len(mutation_order))}
        file_df = pd.DataFrame(file_dict, index = [file_name])
        final_df = pd.concat((final_df, file_df))
    return final_df

def bed_to_counts(bed_path, reference_genome_path):
    # open vcf file
    bed = pd.read_csv(bed_path, header=0, sep='\t', index_col=False)
    # open fasta file
    if reference_genome_path == None:
        print("ERROR: You should provide a path to the reference genome of your data!")
        exit()
    else:
        genome = pysam.FastaFile(reference_genome_path)
    # define by how many bases the variant should be flanked
    flank = 1
    mutation_type_order=os.path.join(DATA, "mutation_type_order.xlsx")
    mutation_order = pd.read_excel(mutation_type_order)
    final_df = pd.DataFrame(columns=mutation_order['Type'])
    list_of_samples = bed['sample'].unique()
    for sample in list_of_samples:
        bed_sample = bed[bed['sample']==sample]
        # iterate over each variant
        list_of_mutations = []
        for index, row in bed_sample.iterrows():
            # extract sequence
            #
            # The start position is calculated by subtract the number of bases
            # given by 'flank' from the variant position. The position in the bed file
            # is 0-based. pysam's fetch() expected 0-base coordinate. 
            #
            # The end position is calculated by adding the number of bases
            # given by 'flank' to the variant position. We also need to add the length
            # of the REF value.
            # 
            # We need to check that 'chr' is contained in the chromosome, otherwise it doesn't work.
            if 'chr' not in row['chr']:
                chr = 'chr' + row['chr']
            else:
                chr = row['chr']
            # Now we have the complete sequence like this:
            # [number of bases given by flank]+REF+[number of bases given by flank]
            seq = genome.fetch(chr, row['start']-flank, row['start']+len(row['ref'])+flank)
            # print out tab seperated columns:
            # CRHOM, POS, REF, ALT, flanking sequencing with variant given in the format '[REF/ALT]'
            mutation = '{}[{}>{}]{}'.format(seq[:flank].upper(), row['ref'].upper(), row['alt'][0].upper(), seq[flank+len(row['ref']):].upper())
            list_of_mutations.append(mutation)
        # Now we count the number of times each mutation type is present in the list
        counts = Counter(list_of_mutations)
        # Filter and sort based on the mutation categories. 
        muts_sorted = []
        for mutation_type in mutation_order['Type']:
            mutation_type_rev = complement(mutation_type[6])+'['+complement(mutation_type[2])+'>'+complement(mutation_type[4])+']'+complement(mutation_type[0])
            muts_sorted.append(counts[mutation_type]+counts[mutation_type_rev])
        sample_dict = {mutation_order.loc[i, 'Type']: [muts_sorted[i]] for i in range(len(mutation_order))}
        sample_df = pd.DataFrame(sample_dict, index = [sample])
        final_df = pd.concat((final_df, sample_df))
    final_df = final_df.rename_axis(None, axis=1)
    return final_df
